# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import torch


__all__ = ["InternalModule"]


class InternalModule(torch.nn.Module):
    """
    Abstract base class for modules which are not a part of the the model definition.
    `torch.nn.Module`s which inherit from this class will not be targeted by configs

    This is typically used to skip apply configs to `Observers` and `Transforms`
    """

    pass



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import gc
import inspect
from weakref import ref

import pytest
import torch
from compressed_tensors.offload import disable_offloading, disable_onloading
from compressed_tensors.offload.cache.cpu import CPUCache
from compressed_tensors.offload.module import offload_module
from tests.test_offload.conftest import assert_device_equal
from tests.testing_utils import requires_gpu


ONLOAD_DEVICE = torch.device("cuda")
OFFLOAD_DEVICE = torch.device("cpu")


@pytest.fixture(scope="function")
def cache():
    return CPUCache(ONLOAD_DEVICE)


@pytest.fixture(scope="function")
def linear():
    return torch.nn.Linear(5, 5, bias=True, device=OFFLOAD_DEVICE)


@pytest.fixture(scope="function")
def offloaded_linear(linear, cache):
    offload_module(linear, ONLOAD_DEVICE, OFFLOAD_DEVICE)
    return linear


@pytest.fixture(scope="function")
def input():
    return torch.zeros(6, device=OFFLOAD_DEVICE)


@pytest.mark.unit
@requires_gpu
def test_onloading(linear: torch.nn.Linear, cache):
    weight = linear.weight
    bias = linear.bias

    offload_module(linear, ONLOAD_DEVICE, OFFLOAD_DEVICE)
    onloaded_weight = linear.weight
    onloaded_bias = linear.bias

    assert_device_equal(onloaded_weight.device, ONLOAD_DEVICE)
    assert_device_equal(onloaded_bias.device, ONLOAD_DEVICE)

    assert type(onloaded_weight) is type(weight)
    assert type(onloaded_bias) is type(bias)
    assert torch.equal(onloaded_weight.to(weight.device), weight)
    assert torch.equal(onloaded_bias.to(bias.device), bias)


@pytest.mark.unit
@requires_gpu
def test_garbage_collect(offloaded_linear: torch.nn.Linear):
    weight_ref = ref(offloaded_linear.weight)
    bias_ref = ref(offloaded_linear.bias)

    del offloaded_linear
    gc.collect()

    assert weight_ref() is None
    assert bias_ref() is None


@pytest.mark.unit
@requires_gpu
def test_disable_offloading(offloaded_linear: torch.nn.Linear):
    outside_onloaded = offloaded_linear.weight
    outside_onloaded_ref = ref(outside_onloaded)
    assert_device_equal(outside_onloaded.device, ONLOAD_DEVICE)

    with disable_offloading():
        inside_onloaded = offloaded_linear.weight
        inside_onloaded_ref = ref(inside_onloaded)
        assert_device_equal(inside_onloaded.device, ONLOAD_DEVICE)

        del outside_onloaded
        del inside_onloaded
        gc.collect()

        assert outside_onloaded_ref() is None
        assert inside_onloaded_ref() is not None

    assert outside_onloaded_ref() is None
    assert inside_onloaded_ref() is None


@pytest.mark.unit
@requires_gpu
def test_disable_onloading(linear: torch.nn.Linear, cache):
    offloaded_weight = linear.weight

    offload_module(linear, ONLOAD_DEVICE, OFFLOAD_DEVICE)

    with disable_onloading():
        weight = linear.weight
        assert weight is offloaded_weight

        # new parameter assignments are direct
        new_param = torch.nn.Parameter(torch.ones(5, device=ONLOAD_DEVICE))
        linear.new_param = new_param
        assert linear.new_param is new_param

    assert weight is offloaded_weight


@pytest.mark.unit
@requires_gpu
def test_delete(offloaded_linear: torch.nn.Linear):
    weight_ref = ref(offloaded_linear.weight)
    bias_ref = ref(offloaded_linear.bias)

    del offloaded_linear.weight
    del offloaded_linear.bias
    gc.collect()

    assert weight_ref() is None
    assert bias_ref() is None


@pytest.mark.unit
@requires_gpu
def test_forward_call(linear: torch.nn.Linear, cache):
    def forward(self, input: torch.Tensor) -> torch.Tensor:
        assert_device_equal(input.device, ONLOAD_DEVICE)
        return torch.nn.functional.linear(input, linear.weight, linear.bias)

    linear.forward = forward.__get__(linear)

    offload_module(linear, ONLOAD_DEVICE, OFFLOAD_DEVICE)

    with torch.no_grad():
        input = torch.zeros(5, device=OFFLOAD_DEVICE)
        output = linear.forward(input)
        assert_device_equal(output.device, ONLOAD_DEVICE)


@pytest.mark.unit
@pytest.mark.parametrize("param_device", (ONLOAD_DEVICE, OFFLOAD_DEVICE))
@pytest.mark.parametrize("use_register_parameter", (True, False))
@pytest.mark.parametrize("requires_grad", (True, False))
def test_register_parameter(
    offloaded_linear: torch.nn.Linear,
    param_device,
    use_register_parameter,
    requires_grad,
):
    # register param
    data = torch.ones(5, device=param_device)
    param = torch.nn.Parameter(data, requires_grad=requires_grad)
    if use_register_parameter:
        offloaded_linear.register_parameter("param_name", param)
    else:
        offloaded_linear.param_name = param

    # new param is correctly onloaded
    assert_device_equal(offloaded_linear.param_name.device, ONLOAD_DEVICE)
    assert torch.equal(offloaded_linear.param_name.to(param_device), param)


@pytest.mark.unit
@pytest.mark.parametrize("param_device", (ONLOAD_DEVICE, OFFLOAD_DEVICE))
@pytest.mark.parametrize("use_register_parameter", (True, False))
@pytest.mark.parametrize("requires_grad", (True, False))
def test_register_parameter_invalidates(
    offloaded_linear: torch.nn.Linear,
    param_device,
    use_register_parameter,
    requires_grad,
):
    with disable_offloading():
        # original weight is kept onloaded
        onloaded_weight = offloaded_linear.weight
        assert onloaded_weight in set(CPUCache.keep_onloaded_values.values())

        # add new param
        data = torch.ones((5, 5), device=param_device)
        param = torch.nn.Parameter(data, requires_grad=requires_grad)
        if use_register_parameter:
            offloaded_linear.register_parameter("weight", param)
        else:
            offloaded_linear.weight = param

        # new param is correct
        assert_device_equal(offloaded_linear.weight.device, ONLOAD_DEVICE)
        assert torch.equal(offloaded_linear.weight, param.to(ONLOAD_DEVICE))


@pytest.mark.unit
def test_forward_signature(linear: torch.nn.Linear, cache):
    original_signature = inspect.signature(linear.forward)

    offload_module(linear, ONLOAD_DEVICE, OFFLOAD_DEVICE)
    assert inspect.signature(linear.forward) == original_signature


@pytest.mark.unit
def test_set_item(offloaded_linear: torch.nn.Linear):
    # update
    update = torch.nn.Parameter(
        torch.rand(5, 5, device=OFFLOAD_DEVICE), requires_grad=False
    )
    offloaded_linear.weight = update
    with disable_onloading():
        assert offloaded_linear.weight is not update

    # overwrite with different size
    overwrite = torch.nn.Parameter(
        torch.rand(6, 6, device=OFFLOAD_DEVICE), requires_grad=False
    )
    offloaded_linear.weight = overwrite
    with disable_onloading():
        assert offloaded_linear.weight is overwrite


@pytest.mark.unit
def test_set_item_buffers(offloaded_linear: torch.nn.Linear):
    # common case: registering buffers of difference sizes twice
    new = torch.rand(5)
    offloaded_linear.register_buffer("buffer", new, persistent=False)
    with disable_onloading():
        assert offloaded_linear.buffer is new

    overwrite = torch.rand(6)
    offloaded_linear.register_buffer("buffer", overwrite, persistent=False)
    with disable_onloading():
        assert offloaded_linear.buffer is overwrite



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import os

from huggingface_hub import list_repo_files
from transformers.file_utils import CONFIG_NAME
from transformers.utils.hub import cached_file


__all__ = [
    "get_checkpoint_files",
    "is_weights_file",
    "find_config_path",
    "find_safetensors_index_path",
]


def is_weights_file(file_name: str) -> bool:
    return any(
        file_name.endswith(suffix)
        for suffix in [
            ".bin",
            ".safetensors",
            ".pth",
            ".msgpack",
            ".pt",
        ]
    )


def get_checkpoint_files(model_stub: str | os.PathLike) -> dict[str, str]:
    # In the future, this function can accept and pass download kwargs to cached_file

    if os.path.exists(model_stub):
        file_paths = _walk_file_paths(model_stub, ignore=".cache")
    else:
        file_paths = list_repo_files(model_stub)

    return {file_path: cached_file(model_stub, file_path) for file_path in file_paths}


def _walk_file_paths(root_dir: str, ignore: str | None = None) -> list[str]:
    """
    Return all file paths relative to the root directory
    """

    all_files = []
    for dirpath, _, filenames in os.walk(root_dir):
        for filename in filenames:
            rel_path = os.path.relpath(os.path.join(dirpath, filename), root_dir)
            if not (ignore and rel_path.startswith(ignore)):
                all_files.append(rel_path)
    return all_files


def find_safetensors_index_path(save_directory: str | os.PathLike) -> str | None:
    for file_name in os.listdir(save_directory):
        if file_name.endswith("safetensors.index.json"):
            return os.path.join(save_directory, file_name)

    return None


def find_config_path(save_directory: str | os.PathLike) -> str | None:
    for file_name in os.listdir(save_directory):
        if file_name in (CONFIG_NAME, "params.json"):
            return os.path.join(save_directory, file_name)

    return None



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import torch
from compressed_tensors.compressors.base import BaseCompressor
from compressed_tensors.compressors.pack_quantized.helpers import (
    pack_to_int32,
    unpack_from_int32,
)
from compressed_tensors.config import CompressionFormat
from compressed_tensors.quantization import (
    QuantizationScheme,
    QuantizationStrategy,
    QuantizationType,
)
from compressed_tensors.quantization.lifecycle.forward import dequantize, quantize
from compressed_tensors.utils import TensorStateDict


__all__ = ["PackedQuantizationCompressor"]


PACK_ZP_STRATS = [
    QuantizationStrategy.GROUP.value,
    QuantizationStrategy.CHANNEL.value,
]


@BaseCompressor.register(name=CompressionFormat.pack_quantized.value)
class PackedQuantizationCompressor(BaseCompressor):
    """
    Compresses a quantized model by packing every eight 4-bit weights into an int32.
    """

    @classmethod
    def compress(
        cls, state_dict: TensorStateDict, scheme: QuantizationScheme
    ) -> TensorStateDict:
        """
        Compress a per-module state dict.

        Quantizes the weight, packs it into int32 as ``weight_packed``, stores
        the original shape as ``weight_shape``, and removes ``weight``. If the
        quantization is asymmetric (GROUP or CHANNEL strategy) the zero-point is
        also packed; otherwise it is dropped.

        :param state_dict: local-name state dict (weight, weight_scale, …)
        :param quantization_args: quantization parameters for the weight
        :param device: device to move compressed tensors to
        :return: compressed state dict
        """
        state_dict = state_dict.copy()
        weight = state_dict.pop("weight")
        scale = state_dict.get("weight_scale")
        zero_point = state_dict.get("weight_zero_point", None)
        g_idx = state_dict.get("weight_g_idx", None)
        weights = scheme.weights

        quantized_weight = quantize(
            x=weight,
            scale=scale,
            zero_point=zero_point,
            g_idx=g_idx,
            args=scheme.weights,
            dtype=torch.int8,
        )
        state_dict["weight_packed"] = pack_to_int32(quantized_weight, weights.num_bits)
        state_dict["weight_shape"] = torch.tensor(weight.shape)

        if not weights.symmetric and weights.strategy in PACK_ZP_STRATS:
            assert zero_point is not None, "Asymmetric quant requires zero-point values"
            packed_zp = pack_to_int32(zero_point, weights.num_bits, packed_dim=0)
            state_dict["weight_zero_point"] = packed_zp.contiguous()

        state_dict = cls._remove_symmetric_zp(state_dict, scheme)

        return state_dict

    @classmethod
    def decompress(
        cls, state_dict: TensorStateDict, scheme: QuantizationScheme
    ) -> TensorStateDict:
        """
        Decompress a per-module state dict.

        Unpacks ``weight_packed`` back to the original weight, removes
        ``weight_packed`` and ``weight_shape``, and unpacks the zero-point
        if present.

        :param state_dict: local-name state dict (weight_packed, weight_scale, …)
        :param quantization_args: quantization parameters for the weight
        :return: decompressed state dict with weight in float dtype
        """
        state_dict = state_dict.copy()
        packed = state_dict.pop("weight_packed")
        scale = state_dict.get("weight_scale")
        zero_point = state_dict.get("weight_zero_point", None)
        g_idx = state_dict.get("weight_g_idx", None)
        original_shape = state_dict.get("weight_shape")
        weights = scheme.weights

        # Unpack zero_point before dequantization if needed
        if not weights.symmetric and weights.strategy in PACK_ZP_STRATS:
            assert zero_point is not None, "Asymmetric quant requires zero-point values"
            original_zp_shape = (*original_shape[:-1], scale.shape[-1])
            zero_point = unpack_from_int32(
                zero_point, weights.num_bits, original_zp_shape, packed_dim=0
            )
            state_dict["weight_zero_point"] = zero_point

        unpacked = unpack_from_int32(packed, weights.num_bits, original_shape)
        state_dict["weight"] = dequantize(
            x_q=unpacked,
            scale=scale,
            zero_point=zero_point,
            g_idx=g_idx,
        )

        return state_dict

    @classmethod
    def can_compress(cls, module_type: type, scheme: QuantizationScheme) -> bool:
        """Pack quantized matches weight-only INT quantization with 4 or 8 bits."""
        return (
            module_type == torch.nn.Linear
            and scheme.weights is not None
            and scheme.input_activations is None
            and scheme.weights.num_bits in (4, 8)
            and scheme.weights.type == QuantizationType.INT.value
        )



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import torch
from compressed_tensors.quantization import QuantizationScheme


class CompressedLinear(torch.nn.Linear):
    """
    Wrapper module for running a compressed forward pass of a quantized Linear module.
    The wrapped layer will decompressed on each forward call.
    """

    @classmethod
    def from_linear(
        cls,
        module: torch.nn.Linear,
        quantization_scheme: QuantizationScheme,
        quantization_format: str,
    ):
        raise ValueError("`CompressedLinear` is no longer supported")



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import pytest
import torch
from compressed_tensors.quantization import (
    FP4_E2M1_DATA,
    FP8_E4M3_DATA,
    QuantizationArgs,
    QuantizationStrategy,
)
from compressed_tensors.quantization.utils import (
    calculate_block_padding,
    calculate_qparams,
    compute_dynamic_scales_and_zp,
    generate_gparam,
    maybe_pad_tensor_for_block_quant,
)


@pytest.mark.parametrize(
    "keepdims,strategy,exp_shape",
    [
        (
            False,
            "tensor",
            torch.Size(
                [
                    1,
                ]
            ),
        ),
        (True, "channel", torch.Size([1, 1])),
        (True, "group", torch.Size([1, 1])),
        (
            False,
            "block",
            torch.Size(
                [
                    1,
                ]
            ),
        ),
    ],
)
def test_calculate_qparams(keepdims, strategy, exp_shape):
    value = torch.empty(5, 6)
    min_val = torch.amin(value, dim=tuple(), keepdims=keepdims)
    max_val = torch.amax(value, dim=tuple(), keepdims=keepdims)

    if strategy == QuantizationStrategy.GROUP:
        args = QuantizationArgs(strategy=strategy, group_size=2)
    elif strategy == QuantizationStrategy.BLOCK:
        args = QuantizationArgs(strategy=strategy, block_structure=[1, 3])
    else:
        args = QuantizationArgs(
            strategy=strategy,
            group_size=(2 if strategy == "group" else None),
            block_structure=([1, 3] if strategy == "block" else None),
        )
        scale, zp = calculate_qparams(min_val, max_val, args)
        assert scale.shape == exp_shape
        assert zp.shape == exp_shape


def test_fused_global_scales():
    layer = torch.nn.Linear(7, 8)
    max_tensor_value = torch.abs(layer.weight.data).max()
    # use defaults
    min_val, max_val = torch.aminmax(layer.weight)
    global_scale = generate_gparam(min_val.data, max_val.data)
    # max value should be = (448 * 6) / global_scale
    assert max_tensor_value.item() == pytest.approx(
        FP4_E2M1_DATA.max * FP8_E4M3_DATA.max / global_scale, abs=0.001
    )


@pytest.mark.parametrize(
    "shape,group_size,exp_shape",
    [
        # Only batch size =1 is supported for dynamic GROUP quantization
        ((1, 4, 8), 4, torch.Size([1, 4, 2])),
    ],
)
def test_compute_dynamic_scales_and_zp_group(shape, group_size, exp_shape):
    """
    Dynamic group quantization should reduce activations in groups, producing
    scales and zero points of shape [batch, num_groups].
    """
    value = torch.randn(*shape)
    args = QuantizationArgs(
        strategy=QuantizationStrategy.GROUP,
        group_size=group_size,
        dynamic=True,
    )
    scale, zp = compute_dynamic_scales_and_zp(value, args, module=torch.nn.Module())
    assert scale.shape == exp_shape
    assert zp.shape == exp_shape


# Tests for block quantization padding utilities


@pytest.mark.parametrize(
    "shape,block_structure,expected_padding",
    [
        # DeepSeek-V2-Lite intermediate_size (10944 % 128 = 64)
        ((10944, 2048), (128, 128), (64, 0)),
        # Both dimensions non-divisible
        ((100, 200), (128, 128), (28, 56)),
        # Only rows non-divisible
        ((300, 256), (128, 128), (84, 0)),
        # Only cols non-divisible
        ((256, 300), (128, 128), (0, 84)),
        # Divisible dimensions (no padding needed)
        ((256, 384), (128, 128), (0, 0)),
        # Smaller than block size
        ((100, 100), (128, 128), (28, 28)),
    ],
)
def test_calculate_block_padding(shape, block_structure, expected_padding):
    """Test that calculate_block_padding computes correct padding amounts."""
    pad_rows, pad_cols = calculate_block_padding(shape, block_structure)
    assert (pad_rows, pad_cols) == expected_padding, (
        f"For shape {shape} with block {block_structure}, "
        f"expected padding {expected_padding}, got ({pad_rows}, {pad_cols})"
    )


@pytest.mark.parametrize(
    "rows,cols,block_height,block_width",
    [
        (10944, 2048, 128, 128),  # DeepSeek-V2-Lite
        (100, 200, 128, 128),  # Both non-divisible
        (256, 256, 128, 128),  # Divisible (no padding)
        (50, 50, 128, 128),  # Smaller than block
    ],
)
def test_maybe_pad_tensor_for_block_quant(rows, cols, block_height, block_width):
    """Test that maybe_pad_tensor_for_block_quant correctly pads tensors."""
    tensor = torch.randn(rows, cols)
    block_structure = (block_height, block_width)

    padded = maybe_pad_tensor_for_block_quant(tensor, block_structure)

    # Check padded dimensions are divisible by block size
    assert (
        padded.shape[-2] % block_height == 0
    ), f"Padded rows {padded.shape[-2]} should be divisible by {block_height}"
    assert (
        padded.shape[-1] % block_width == 0
    ), f"Padded cols {padded.shape[-1]} should be divisible by {block_width}"

    # Check that original values are preserved
    assert torch.equal(
        padded[:rows, :cols], tensor
    ), "Original values should be preserved in padded tensor"

    # Check that padding is zeros
    if padded.shape[-2] > rows:
        assert torch.all(padded[rows:, :] == 0), "Row padding should be zeros"
    if padded.shape[-1] > cols:
        assert torch.all(padded[:, cols:] == 0), "Column padding should be zeros"



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import pytest
import torch
from compressed_tensors.offload import (
    align_module_device,
    align_modules,
    disable_offloading,
    disable_onloading,
    get_execution_device,
    get_offloaded_device,
    update_offload_parameter,
)
from compressed_tensors.offload.cache import CPUCache
from compressed_tensors.offload.module import offload_module
from tests.test_offload.conftest import assert_device_equal, assert_tensor_equal
from tests.testing_utils import requires_gpu


ONLOAD_DEVICE = torch.device("cuda")
OFFLOAD_DEVICE = torch.device("cpu")


@pytest.fixture(scope="function")
def cache():
    return CPUCache(ONLOAD_DEVICE)


@pytest.fixture(scope="function")
def linear():
    return torch.nn.Linear(5, 5, bias=True, device=OFFLOAD_DEVICE)


@pytest.fixture(scope="function")
def offloaded_linear(linear, cache):
    offload_module(linear, ONLOAD_DEVICE, OFFLOAD_DEVICE)
    return linear


@pytest.mark.unit
@requires_gpu
def test_disable_offloading():
    cache1 = CPUCache(ONLOAD_DEVICE)
    cache2 = CPUCache(ONLOAD_DEVICE)

    cache1["weight"] = torch.tensor(0, device=OFFLOAD_DEVICE)
    cache2["weight"] = torch.tensor(1, device=OFFLOAD_DEVICE)

    with disable_offloading():
        assert cache1["weight"] in cache1.keep_onloaded_values.values()
        assert cache2["weight"] in cache2.keep_onloaded_values.values()


@pytest.mark.unit
@requires_gpu
def test_disable_onloading():
    cache1 = CPUCache(ONLOAD_DEVICE)
    cache2 = CPUCache(ONLOAD_DEVICE)

    cache1["weight"] = torch.tensor(0, device=OFFLOAD_DEVICE)
    cache2["weight"] = torch.tensor(1, device=OFFLOAD_DEVICE)

    with disable_onloading():
        assert_device_equal(cache1["weight"].device, OFFLOAD_DEVICE)
        assert_device_equal(cache2["weight"].device, OFFLOAD_DEVICE)


@pytest.mark.unit
@requires_gpu
@pytest.mark.parametrize("offload", (True, False))
def test_update_offload_parameter(linear: torch.nn.Linear, cache, offload):
    init_data = torch.tensor(0.0, device=OFFLOAD_DEVICE)
    linear.weight = torch.nn.Parameter(init_data, requires_grad=False)
    if offload:
        offload_module(linear, ONLOAD_DEVICE, OFFLOAD_DEVICE)

    assert linear.weight == 0

    update_offload_parameter(linear, "weight", torch.tensor(1))
    assert linear.weight == 1

    with disable_offloading():
        update_offload_parameter(linear, "weight", torch.tensor(2))
        assert linear.weight == 2
    assert linear.weight == 2

    with disable_onloading():
        update_offload_parameter(linear, "weight", torch.tensor(3))
        assert linear.weight == 3
    assert linear.weight == 3


@pytest.mark.unit
def test_update_offload_parameter_with_grad(linear: torch.nn.Linear):
    zeros = torch.nn.Parameter(torch.zeros(5, 5), requires_grad=True)
    update_offload_parameter(linear, "weight", zeros)
    assert_tensor_equal(linear.weight, zeros)

    ones = torch.nn.Parameter(torch.ones(5, 5), requires_grad=True)
    offload_module(linear, ONLOAD_DEVICE, OFFLOAD_DEVICE)
    update_offload_parameter(linear, "weight", ones)
    assert_tensor_equal(linear.weight, ones, ONLOAD_DEVICE)


@pytest.mark.unit
@requires_gpu
def test_get_execution_device(linear: torch.nn.Linear, cache):
    assert_device_equal(get_execution_device(linear), OFFLOAD_DEVICE)
    linear.to(ONLOAD_DEVICE)
    assert_device_equal(get_execution_device(linear), ONLOAD_DEVICE)

    linear.to(OFFLOAD_DEVICE)
    offload_module(linear, ONLOAD_DEVICE, OFFLOAD_DEVICE)
    assert_device_equal(get_execution_device(linear), ONLOAD_DEVICE)

    with disable_onloading():
        assert_device_equal(get_execution_device(linear), ONLOAD_DEVICE)

    with disable_offloading():
        assert_device_equal(get_execution_device(linear), ONLOAD_DEVICE)


@pytest.mark.unit
@requires_gpu
def test_get_offloaded_device(linear: torch.nn.Linear, cache):
    assert_device_equal(get_offloaded_device(linear), OFFLOAD_DEVICE)
    linear.to(ONLOAD_DEVICE)
    assert_device_equal(get_offloaded_device(linear), ONLOAD_DEVICE)

    linear.to(OFFLOAD_DEVICE)
    offload_module(linear, ONLOAD_DEVICE, OFFLOAD_DEVICE)
    assert_device_equal(get_offloaded_device(linear), OFFLOAD_DEVICE)

    with disable_onloading():
        assert_device_equal(get_offloaded_device(linear), OFFLOAD_DEVICE)

    with disable_offloading():
        assert_device_equal(get_offloaded_device(linear), OFFLOAD_DEVICE)


@pytest.mark.unit
@requires_gpu
def register_offload_module(linear: torch.nn.Linear, cache):
    sub1 = torch.nn.Linear(1, 1)
    register_offload_module(linear, "sub1", sub1)
    assert linear.sub1 is sub1

    offload_module(linear, ONLOAD_DEVICE, OFFLOAD_DEVICE)
    sub2 = torch.nn.Linear(1, 1)
    register_offload_module(linear, "sub2", sub2)
    assert linear.sub2 is sub2
    assert_device_equal(sub2.weight.device, ONLOAD_DEVICE)


@pytest.mark.unit
@requires_gpu
def test_align_modules(offloaded_linear: torch.nn.Linear):
    linear = torch.nn.Linear(1, 1, device=ONLOAD_DEVICE)

    with align_modules((linear, offloaded_linear), OFFLOAD_DEVICE):
        assert_device_equal(linear.weight.device, OFFLOAD_DEVICE)
        assert_device_equal(offloaded_linear.weight.device, OFFLOAD_DEVICE)


@pytest.mark.unit
@requires_gpu
@pytest.mark.parametrize("offload", (True, False))
def test_align_module_device(linear: torch.nn.Linear, cache, offload):
    if offload:
        offload_module(linear, ONLOAD_DEVICE, OFFLOAD_DEVICE)
    else:
        linear.to(ONLOAD_DEVICE)

    with align_module_device(linear, OFFLOAD_DEVICE):
        assert_device_equal(linear.weight.device, OFFLOAD_DEVICE)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import pytest
from compressed_tensors.config import SparsityStructure


def test_sparsity_structure_valid_cases():
    assert (
        SparsityStructure("2:4") == SparsityStructure.TWO_FOUR
    ), "Failed to match '2:4' with TWO_FOUR"
    assert (
        SparsityStructure("unstructured") == SparsityStructure.UNSTRUCTURED
    ), "Failed to match 'unstructured' with UNSTRUCTURED"
    assert (
        SparsityStructure("UNSTRUCTURED") == SparsityStructure.UNSTRUCTURED
    ), "Failed to match 'UNSTRUCTURED' with UNSTRUCTURED"
    assert (
        SparsityStructure(None) == SparsityStructure.UNSTRUCTURED
    ), "Failed to match None with UNSTRUCTURED"


def test_sparsity_structure_invalid_case():
    with pytest.raises(ValueError, match="invalid is not a valid SparsityStructure"):
        SparsityStructure("invalid")


def test_sparsity_structure_case_insensitivity():
    assert (
        SparsityStructure("2:4") == SparsityStructure.TWO_FOUR
    ), "Failed to match '2:4' with TWO_FOUR"
    assert (
        SparsityStructure("2:4".upper()) == SparsityStructure.TWO_FOUR
    ), "Failed to match '2:4'.upper() with TWO_FOUR"
    assert (
        SparsityStructure("unstructured".upper()) == SparsityStructure.UNSTRUCTURED
    ), "Failed to match 'unstructured'.upper() with UNSTRUCTURED"
    assert (
        SparsityStructure("UNSTRUCTURED".lower()) == SparsityStructure.UNSTRUCTURED
    ), "Failed to match 'UNSTRUCTURED'.lower() with UNSTRUCTURED"


def test_sparsity_structure_default_case():
    assert (
        SparsityStructure(None) == SparsityStructure.UNSTRUCTURED
    ), "Failed to match None with UNSTRUCTURED"



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from compressed_tensors.transform import TransformArgs, TransformScheme


def test_basic_scheme():
    targets = ["Linear"]
    location = "input"
    basic_args = TransformArgs(targets=targets, location=location)

    scheme = TransformScheme(
        type="hadamard",
        apply=[basic_args],
    )
    assert not scheme.randomize
    assert scheme.type == "hadamard"
    assert len(scheme.apply) == 1
    assert isinstance(scheme.apply[0], TransformArgs)


def test_multiple_groups_global():
    targets = ["Embedding"]
    location = "input"
    embedding_args = TransformArgs(targets=targets, location=location)

    targets = ["Linear"]
    location = "weight_input"
    linear_args = TransformArgs(targets=targets, location=location)

    # same transform applied to multiple groups
    scheme = TransformScheme(
        type="hadamard",
        apply=[embedding_args, linear_args],
        randomize=True,
    )

    assert scheme.randomize
    assert scheme.type == "hadamard"
    assert len(scheme.apply) == 2
    assert isinstance(scheme.apply[0], TransformArgs)
    assert isinstance(scheme.apply[1], TransformArgs)


def test_multiple_groups():
    apply = []
    location = "weight_output"

    for i in range(20):
        targets = [f"model.layers.{i}.attn.v_proj", f"model.layers.{i}.attn.o_proj"]
        args = TransformArgs(targets=targets, location=location)
        apply.append(args)

    # global is False, different hadamard transform applied to each group
    # same dimension/hidden dim
    scheme = TransformScheme(
        type="hadamard",
        apply=apply,
    )

    assert not scheme.randomize
    assert scheme.type == "hadamard"
    assert len(scheme.apply) == 20



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from collections.abc import Container
from copy import deepcopy
from functools import partial
from typing import Any, Optional, TypeVar

import torch
import torch.distributed as dist
from compressed_tensors.offload.cache import OffloadCache
from compressed_tensors.offload.module import offload_module, remove_module_offload
from compressed_tensors.offload.utils import (
    get_module_device,
    get_module_sizes,
    module_size,
)
from compressed_tensors.utils import getattr_chain
from compressed_tensors.utils.binary_search import SearchFailureError, max_binary_search
from loguru import logger
from transformers import PreTrainedModel


__all__ = [
    "offload_model",
    "dispatch_with_map",
    "get_device_map",
    "dispatch_model",
    "remove_dispatch",
    "get_device_memory",
    "DeviceMap",
]

ModelType = TypeVar("ModelType", bound=torch.nn.Module)
DeviceMap = dict[str, tuple[torch.device | None, torch.device | str | None]]


def offload_model(
    model: ModelType,
    onload_device: torch.device | str,
    offload_device: Any = None,
) -> ModelType:
    """
    Modify the dispatch of a model to onload to the provided `onload_device`. Existing
    offloaded tensors will not be modified. If a module is not offloaded, it will be
    offloaded to the provided `offload_device`.

    :param model: model to dispatch
    :param onload_device: device to move weights to during forward pass
    :param offload_device: device to offload weights to, if not already offloaded
    :return: dispatched model
    """
    if offload_device is not None:
        logger.warning(
            "`offload_model` now keeps the same offload device that model was loaded "
            "on. Please specify offload by loading the model on its offload device(s)"
        )

    # offload modules in place
    for module in model.modules():
        if isinstance(module._parameters, OffloadCache):
            module._parameters.onload_device = onload_device
            module._buffers.onload_device = onload_device
        else:
            offload_device = get_module_device(module, torch.device("cpu"))
            offload_module(module, onload_device, offload_device)

    return model


def dispatch_with_map(
    model: torch.nn.Module,
    device_map: DeviceMap,
    offload_dir: Optional[str] = None,
):
    """
    Dispatch a model according to the provided device map

    :param model: model to dispatch
    :param device_map: device map specifying the onload and offload of each module
    :param offload_dir: optional directory for disk offloading
    """
    for name, (onload_device, offload_device) in device_map.items():
        module = model.get_submodule(name)

        if offload_device == "disk":
            offload_module(
                module, onload_device, offload_device, offload_dir=offload_dir
            )

        elif offload_device is not None:
            offload_module(module, onload_device, offload_device)


def get_device_map(
    model: torch.nn.Module, default_device: torch.device = torch.device("cpu")
) -> DeviceMap:
    """
    Get the device map of a CT-offloaded model

    :param: model: model to get device map of
    :param default_device: the default onload/offload device
        when module has no parameters
    :return: device map specifying the onload and offload device of all modules
    """
    from compressed_tensors.offload import get_execution_device, get_offloaded_device

    return {
        name: (
            get_execution_device(module, default_device),
            get_offloaded_device(module, default_device),
        )
        for name, module in model.named_modules(remove_duplicate=False)
    }


def dispatch_model(
    model: ModelType,
    device_memory: dict[torch.device, int] | None = None,
    extra_memory: int | None = None,
    no_split_modules: Container[str] | None = None,
) -> ModelType:
    """
    Dispatch a model for autoregressive generation. This means that modules are
    dispatched evenly across available devices and kept onloaded if possible. If
    onloading the entire model is not possible, some modules may be offloaded. Any
    existing offloads will be removed.

    Disclaimers:
    * Optimal runtime assumes that modules are called in order of `model.modules()`

    :param model: model to dispatch
    :param device_memory: optional dictionary mapping torch device to available memory.
        If none is provided, all available devices will be used
    :param extra_memory: the amount of memory to be reserved for activations
    :param no_split_modules: names of module classes which should not be split
        across multiple devices
    :return: dispatched model
    """
    # infer no_split_modules
    if no_split_modules is None:
        no_split_modules = getattr(model, "_no_split_modules", tuple())

    # collect devices
    if device_memory is None:
        device_memory: dict[torch.device, int] = get_device_memory()
    if len(device_memory) <= 0:
        raise MemoryError("Did not find any devices to dispatch model to")

    # collect module sizes
    sizes = get_module_sizes(model, no_split_modules)
    if len(sizes) <= 0:
        raise ValueError("Model does not have any modules")

    # estimate memory requirement
    if extra_memory is None:
        # fragmentation, kv cache, embeddings, ect.
        extra_memory = max(module_size(model) * 0.05, 1e9)

        # activations
        if isinstance(model, PreTrainedModel):
            extra_memory += (
                1  # batch_size
                * 2048  # seq_len
                * getattr_chain(model, "config.intermediate_size", 256)
                * getattr(model, "dtype", torch.bfloat16).itemsize
            )

    # search for the best dispatch which maximizes extra memory across devices
    try:
        max_extra_memory = min(device_memory.values())
        extra_memory, (dispatch, _) = max_binary_search(
            fn=partial(_get_greedy_dispatch, sizes, device_memory),
            cond=(lambda result: len(result[0]) == len(sizes)),
            start=extra_memory,
            end=max_extra_memory,
        )

    # fallback: create a cpu dispatch
    except SearchFailureError:
        dispatch, device_memory = _get_greedy_dispatch(
            sizes, device_memory, extra_memory
        )
        assert len(dispatch) < len(sizes)

        last_device = dispatch[-1][1] if len(dispatch) else list(device_memory)[0]
        sizes_dict = {module: size for module, size in sizes}
        largest_offloaded_module = max(size for _, size in sizes[len(dispatch) :])

        # pop off modules until all offloaded modules can fit in last device
        while largest_offloaded_module + extra_memory > device_memory[last_device]:
            if len(dispatch) <= 0:
                raise ValueError(
                    f"Cannot fit no_split module of size {largest_offloaded_module} "
                    f"bytes into any device: {device_memory}"
                )

            module, last_device, _ = dispatch.pop(-1)
            device_memory[last_device] += sizes_dict[module]
            largest_offloaded_module = max(largest_offloaded_module, sizes_dict[module])

        # fill dispatch back with cpu offloading
        for module, _ in list(sizes[len(dispatch) :]):
            dispatch.append((module, last_device, "cpu"))

        logger.warning("Forced to offload modules due to insufficient gpu resources")

    # dispatch
    assert len(dispatch) == len(sizes)

    dispatch_dict = {
        submodule: (onload, offload)
        for module, onload, offload in dispatch
        for submodule in module.modules()
    }

    for module in model.modules():
        remove_module_offload(module, onload_tensors=True)
        if module in dispatch_dict:
            onload, offload = dispatch_dict[module]
            offload_module(module, onload, offload)

    logger.debug(f"Dispatched model with {extra_memory} bytes of extra memory")
    return model


def get_device_memory() -> dict[torch.device, int]:
    """
    Get the total memory of all available devices. Returns CUDA device memory
    when available, otherwise falls back to CPU with system RAM.

    :return: mapping from torch device to total memory
    """
    if not torch.cuda.is_available():
        import os

        total_ram = os.sysconf("SC_PAGE_SIZE") * os.sysconf("SC_PHYS_PAGES")
        return {torch.device("cpu"): total_ram}

    if dist.is_available() and dist.is_initialized():
        logger.info("Detected distributed context. Dispatching to local rank gpu")
        device_memory = torch.cuda.get_device_properties(dist.get_rank()).total_memory
        return {torch.device("cuda"): device_memory}

    return {
        # TODO: extend to xpu, ect.
        torch.device(f"cuda:{idx}"): torch.cuda.get_device_properties(idx).total_memory
        for idx in range(torch.cuda.device_count())
    }


def remove_dispatch(
    module: torch.nn.Module, onload_tensors: bool = False
) -> torch.nn.Module:
    """
    Remove any existing dispatches from module

    :param onload_tensors: Whether to move tensors to the onloaded device, or keep them
        on the offload device. Defaults to False.
    :return: module with offloading functionality removed
    """
    for submodule in module.modules():
        remove_module_offload(submodule, onload_tensors)

    return module


def _get_greedy_dispatch(
    sizes: list[tuple[torch.nn.Module, int]],
    device_memory: dict[torch.device, int],
    extra_memory: int = 0,
) -> tuple[
    list[tuple[torch.nn.Module, torch.device, torch.device]], dict[torch.device, int]
]:
    dispatch = list()
    memory_remaining = deepcopy(device_memory)

    device_index = 0
    devices = list(memory_remaining.keys())

    if len(devices) <= 0:
        raise ValueError()

    for module, size in sizes:
        while True:
            if device_index >= len(devices):
                return dispatch, memory_remaining

            device = devices[device_index]
            if size > memory_remaining[device] - extra_memory:
                device_index += 1
                continue

            dispatch.append((module, device, device))
            memory_remaining[device] -= size
            break

    return dispatch, memory_remaining



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from abc import ABC
from typing import Optional

import torch
from compressed_tensors.compressors.format import infer_module_format
from compressed_tensors.config import CompressionFormat
from compressed_tensors.quantization import QuantizationScheme, QuantizationStatus
from compressed_tensors.registry import RegistryMixin
from compressed_tensors.utils import (
    TensorStateDict,
    get_direct_state_dict,
    replace_direct_state_dict,
)


__all__ = ["BaseCompressor", "compress_module", "decompress_module"]


class BaseCompressor(RegistryMixin, ABC):
    """
    Base class representing a model compression algorithm.

    New quantization compressors (dense, naive_quantized, pack_quantized, nvfp4,
    mxfp4) use the classmethod interface — they are never instantiated. Look up
    via BaseCompressor.get_value_from_registry(format) and call compress/decompress
    directly on the returned class.

    Legacy sparse compressors (sparse_bitmask, sparse_24_bitmask, marlin_24) still
    use the instance-based interface and are instantiated via load_from_registry.
    """

    @classmethod
    def compress(
        cls, state_dict: TensorStateDict, scheme: QuantizationScheme
    ) -> TensorStateDict:
        """
        Compress a per-module state dict. Does not modify the input.

        Keys are *local* names (``weight``, ``weight_scale``, …), not prefixed
        with the module path.

        :param state_dict: per-module state dict with local parameter names
        :param scheme: quantization scheme containing quantization parameters
        :return: compressed per-module state dict
        """
        raise NotImplementedError(
            f"{cls.__name__} does not implement the classmethod compress interface"
        )

    @classmethod
    def decompress(
        cls, state_dict: TensorStateDict, scheme: QuantizationScheme
    ) -> TensorStateDict:
        """
        Decompress a per-module state dict. Does not modify the input.

        Keys are *local* names (``weight_packed``, ``weight_scale``, …).

        :param state_dict: compressed per-module state dict with local parameter names
        :param scheme: quantization scheme containing quantization parameters
        :return: decompressed per-module state dict
        """
        raise NotImplementedError(
            f"{cls.__name__} does not implement the classmethod decompress interface"
        )

    @classmethod
    def compress_module(cls, module: torch.nn.Module) -> None:
        """
        Compress a module in-place by compressing its state dict.

        Extracts the module's parameters and buffers, compresses them using the
        compress classmethod, and replaces the module's state with the compressed
        version.

        :param module: the module to compress in-place
        """
        scheme = getattr(module, "quantization_scheme")

        state_dict = get_direct_state_dict(module)
        compressed_state_dict = cls.compress(state_dict, scheme)
        replace_direct_state_dict(module, compressed_state_dict)

        module.quantization_status = QuantizationStatus.COMPRESSED

    @classmethod
    def decompress_module(cls, module: torch.nn.Module) -> None:
        """
        Decompress a module in-place by decompressing its state dict.

        Extracts the module's parameters and buffers, decompresses them using the
        decompress classmethod, and replaces the module's state with the decompressed
        version.

        :param module: the module to decompress in-place
        """
        scheme = getattr(module, "quantization_scheme")

        state_dict = get_direct_state_dict(module)
        decompressed_state_dict = cls.decompress(state_dict, scheme)
        replace_direct_state_dict(module, decompressed_state_dict)

        module.quantization_status = QuantizationStatus.DECOMPRESSED

    @classmethod
    def can_compress(cls, module_type: type, scheme: QuantizationScheme) -> bool:
        """
        Determine if this compressor is applicable for the given module type and scheme.

        Examines the module type and quantization scheme and determines whether this
        compressor can handle the module's compression requirements.

        :param module_type: the type of the module to check for compatibility
        :param scheme: the quantization scheme to check for compatibility
        :return: True if this compressor can handle the module, False otherwise
        """
        raise NotImplementedError(f"{cls.__name__} does not implement match")

    @classmethod
    def _remove_symmetric_zp(
        cls, state_dict: TensorStateDict, scheme: QuantizationScheme
    ) -> TensorStateDict:
        """
        Remove zero points where quantization arguments do not specify asymmetric quant.
        This is required during compression because vLLM does not support loading
        zero point parameters for symmetric schemes.
        TODO: remove zero points from initialization

        :param state_dict: state dict containing extra zero point parameters
        :return: state dict with extra zero point parameters removed
        """
        if scheme.input_activations and scheme.input_activations.symmetric:
            state_dict.pop("input_zero_point", None)
        if scheme.weights and scheme.weights.symmetric:
            state_dict.pop("weight_zero_point", None)
        if scheme.output_activations and scheme.output_activations.symmetric:
            state_dict.pop("output_zero_point", None)

        return state_dict


def compress_module(
    module: torch.nn.Module, format: Optional[CompressionFormat] = None
):
    """
    Compress a module which has had quantization applied to it. Sets the
    module's quantization format attribute to whichever format was used to compress.

    The format used to compress will be found from one of the following locations:
    1. the `format` argument passed to this function
    2. the attached `quantization_scheme.format` attribute
    3. format inferred from `infer_module_format`

    :param module: module to compress inplace
    :param format: force override for compression format
    """
    scheme = getattr(module, "quantization_scheme", None)
    if not isinstance(scheme, QuantizationScheme):
        return

    scheme.format = CompressionFormat(
        format or scheme.format or infer_module_format(type(module), scheme)
    )
    compressor = BaseCompressor.get_value_from_registry(scheme.format.value)
    compressor.compress_module(module)


def decompress_module(
    module: torch.nn.Module, format: Optional[CompressionFormat] = None
):
    """
    Decompress a module which has had quantization applied to it. Sets the
    module's quantization format attribute to whichever format was used to decompress.

    The format used to decompress will be found from one of the following locations:
    1. the `format` argument passed to this function
    2. the attached `quantization_scheme.format` attribute
    3. format inferred from `infer_module_format`

    :param module: module to decompress inplace
    :param format: force override for decompression format
    """
    scheme = getattr(module, "quantization_scheme", None)
    if not isinstance(scheme, QuantizationScheme):
        return

    scheme.format = CompressionFormat(
        format or scheme.format or infer_module_format(type(module), scheme)
    )
    compressor = BaseCompressor.get_value_from_registry(scheme.format.value)
    compressor.decompress_module(module)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

# flake8: noqa
# isort: skip_file

from .transform_args import *
from .transform_scheme import *
from .transform_config import *

from .factory.base import *
from .factory.hadamard import *
from .factory.matrix_multiply import *
from .factory.random_hadamard import *
from .apply import *



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from unittest.mock import MagicMock, patch

import pytest
import torch
from compressed_tensors.offload import (
    disable_onloading,
    from_accelerate,
    get_offloaded_device,
)
from compressed_tensors.offload.convert import to_accelerate
from compressed_tensors.offload.convert.from_accelerate import _infer_module_device
from compressed_tensors.offload.load import load_offloaded_model, patch_from_pretrained
from tests.test_offload.conftest import assert_device_equal, torchrun
from tests.testing_utils import requires_gpu
from transformers import AutoModelForCausalLM


acclerate = pytest.importorskip("accelerate")


TEST_PARAMETERS = [
    (
        "auto",
        {0: 596049920, "cpu": 1e15},  # force cpu offload for testing
        torch.device("cuda"),
        torch.device("cpu"),
    ),
    (
        "cuda",
        None,
        torch.device("cuda"),
        torch.device("cuda"),
    ),
    (
        "cpu",
        None,
        torch.device("cpu"),
        torch.device("cpu"),
    ),
    (
        "auto_offload",
        {"cpu": 596049920},  # force disk offload for testing
        torch.device("cpu"),
        "disk",
    ),
]


@pytest.mark.integration
@requires_gpu
@pytest.mark.parametrize("device_map,max_memory,first,second", TEST_PARAMETERS)
def test_load(device_map, max_memory, first, second, tmp_path):
    with load_offloaded_model():
        model = AutoModelForCausalLM.from_pretrained(
            "Qwen/Qwen3-0.6B",
            device_map=device_map,
            max_memory=max_memory,
            dtype=torch.bfloat16,
            offload_folder=str(tmp_path / "disk_offload"),
        )

    for layer_index in range(0, 8):
        module = model.get_submodule(f"model.layers.{layer_index}.self_attn.q_proj")
        assert_device_equal(get_offloaded_device(module), first)

    for layer_index in range(8, 28):
        module = model.get_submodule(f"model.layers.{layer_index}.self_attn.q_proj")
        assert_device_equal(get_offloaded_device(module), second)

    with disable_onloading():
        state_dict = model.state_dict(keep_vars=True)

    to_accelerate(model)

    for layer_index in range(0, 8):
        module = model.get_submodule(f"model.layers.{layer_index}.self_attn.q_proj")
        assert_device_equal(_get_accelerate_offloaded_device(module), first)

    for layer_index in range(8, 28):
        module = model.get_submodule(f"model.layers.{layer_index}.self_attn.q_proj")
        assert_device_equal(_get_accelerate_offloaded_device(module), second)

    model.save_pretrained(tmp_path / "save_path")

    from_accelerate(model)

    # TODO: accelerate's disk onloading implementation does not keep consistent meta
    # tensors, :. tensor pointers change and cannot be converted back properly
    if second != "disk":
        with disable_onloading():
            assert model.state_dict(keep_vars=True) == state_dict


@pytest.mark.integration
@requires_gpu(2)
@torchrun(world_size=2)
def test_load_dist(tmp_path):
    for parameters in TEST_PARAMETERS:
        test_load(*parameters, tmp_path=tmp_path)


def _get_accelerate_offloaded_device(module: torch.nn.Module) -> str | None:
    device = _infer_module_device(module)
    if device == torch.device("meta"):
        return "disk"

    return device


@pytest.mark.unit
@patch("compressed_tensors.offload.load.from_accelerate")
@patch("compressed_tensors.offload.load.is_rank0", return_value=True)
@patch("compressed_tensors.offload.load.is_distributed", return_value=False)
def test_patch_forwards_positional_args(
    mock_distributed, mock_rank0, mock_from_accelerate
):
    """Regression: positional args must be forwarded without rebinding to cls."""
    received = {}

    class FakeModel:
        @classmethod
        def from_pretrained(cls, pretrained_model_name_or_path, *model_args, **kwargs):
            received["cls"] = cls
            received["path"] = pretrained_model_name_or_path
            received["model_args"] = model_args
            received["kwargs"] = kwargs
            return MagicMock()

    with patch_from_pretrained(FakeModel, extra_cpu_mem=0):
        FakeModel.from_pretrained(
            "org/model-name", device_map="cpu", torch_dtype="auto"
        )

    assert received["cls"] is FakeModel
    assert received["path"] == "org/model-name"
    assert received["kwargs"]["device_map"] == "cpu"
    assert received["kwargs"]["torch_dtype"] == "auto"



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import contextlib
from abc import ABC, abstractmethod
from collections.abc import Hashable, MutableMapping
from typing import ClassVar, Literal

import torch
import torch.distributed as dist


class OffloadCache(MutableMapping, ABC):
    """
    Base class for offload caches. Subclasses must implement `offload` and `onload`.
    Instances have similar behavior to dicts, except that tensors are offloaded when
    assigned and onloaded when accessed.

    Typical usage:
    ```
    module._parameters = cache_cls.from_mapping(
        module._parameters, onload_device, offload_device
    )
    tensor = ...
    module._parameters["name"] = tensor           # tensor is offloaded
    onloaded_tensor = module._parameters["name"]  # tensor is onloaded
    ```

    This class implements two contexts for more fine-grained control of device movement:
    `OffloadCache.disable_offloading` and `OffloadCache.disable_onloading`. For more
    info, see `compressed_tensors.offload::(disable_offloading|disable_onloading)`
    """

    onload_device: torch.device | str
    offload_device: torch.device | Literal["disk"]

    # global flags for disabling
    offloading_disabled: ClassVar[bool] = False
    onloading_disabled: ClassVar[bool] = False

    # names -> offloaded tensors (populated from _parameters or _buffers)
    offloaded_values: dict[Hashable, torch.Tensor]

    # offloaded tensors -> onloaded tensors (only when offloading is disabled)
    keep_onloaded_values: ClassVar[dict[torch.Tensor, torch.Tensor]] = dict()

    @classmethod
    def cls_from_device(
        cls,
        device: torch.device | str | Literal["disk"] | None = None,
    ) -> type["OffloadCache"]:
        """
        Get the subclass which implements offloading for the given `offload_device`.
        Use `torch.distributed` to detect if the environment is distributed

        :param device: offload device used to find subclass
        :return: subclass of `OffloadCache`
        """
        from compressed_tensors.offload.cache.cpu import CPUCache
        from compressed_tensors.offload.cache.device import DeviceCache
        from compressed_tensors.offload.cache.disk import DiskCache
        from compressed_tensors.offload.cache.dist_cpu import DistributedCPUCache
        from compressed_tensors.offload.cache.dist_device import DistributedDeviceCache
        from compressed_tensors.offload.cache.dist_disk import DistributedDiskCache

        device_type = torch.device(device).type if device != "disk" else "disk"
        distributed = dist.is_available() and dist.is_initialized()

        match (device_type, distributed):
            case ("cpu", False):
                return CPUCache
            case ("cpu", True):
                return DistributedCPUCache
            case ("cuda", False):
                return DeviceCache
            case ("cuda", True):
                return DistributedDeviceCache
            case ("disk", False):
                return DiskCache
            case ("disk", True):
                return DistributedDiskCache
            case _:
                raise NotImplementedError(
                    f"Offload of type {device} and "
                    f"distributed={distributed} has not been implemented"
                )

    @classmethod
    def from_mapping(
        cls,
        mapping: MutableMapping[Hashable, torch.Tensor | None],
        onload_device: torch.device | str,
        offload_device: "torch.device | str | Literal['disk'] | None" = None,
        **kwargs,
    ):
        """
        Initialize an instance from a given mapping, typically `Module._parameters` or
        `Module._buffers`. Mapping values will be offloaded

        :param mapping: mapping used to populate cache
        :param onload_device: device which tensors will be onloaded to
        :param offload_device: device to offload tensors to. For DeviceCache, this
            sets the offload target (defaults to onload_device if not provided).
            For CPUCache and DiskCache, this is validated against the fixed
            offload_device if provided.
        :param \\**kwargs: keyword arguments for cache constructor
        """
        instance = cls(
            onload_device=onload_device, offload_device=offload_device, **kwargs
        )
        instance.offloaded_values = {
            name: instance.offload(tensor) for name, tensor in mapping.items()
        }

        return instance

    def __init__(
        self,
        onload_device: torch.device | str,
        offload_device: torch.device | str | Literal["disk"] | None = None,
    ):
        super().__init__()
        self.onload_device = onload_device
        self.offloaded_values = dict()

        # Validate offload_device for subclasses with a fixed offload_device
        # (CPUCache, DiskCache). DeviceCache sets offload_device after super().__init__
        # so this check only applies when offload_device is a class attribute.
        if offload_device is not None and hasattr(type(self), "offload_device"):
            assert str(offload_device) == str(self.offload_device)

    @abstractmethod
    def onload(self, offloaded: torch.Tensor | None) -> torch.Tensor | None:
        """
        Given an offloaded tensor, returns that tensor after onloading

        :param offloaded: offloaded tensor
        :return: onloaded tensor
        """
        raise NotImplementedError()

    @abstractmethod
    def offload(self, tensor: torch.Tensor | None) -> torch.Tensor | None:
        """
        Given a tensor, returns that tensor after offloading

        :param tensor: tensor to offload
        :return: offloaded tensor
        """
        raise NotImplementedError()

    @abstractmethod
    def update_offload(self, offloaded: torch.Tensor, data: torch.Tensor | None):
        """
        Update the data of an offloaded tensor

        NOTE: Operation is performed asynchronously. If you need the offloaded value
        to update across all ranks, call `dist.barrier()` after calling this function

        :param tensor: offloaded tensor to update
        :param data: new tensor data
        """
        raise NotImplementedError()

    def __getitem__(self, key: Hashable) -> torch.Tensor:
        """
        Onload a tensor

        If called within the `disable_offloading` context, a strong reference of the
        onloaded tensor is kept so that future accesses will not require device movement

        :param key: name of tensor to access
        :return: onloaded tensor
        """
        offloaded = self.offloaded_values[key]

        # when onloading is disabled, offloaded tensors can be accessed directly
        if offloaded is None or self.onloading_disabled:
            return offloaded

        # check for cache hit
        if offloaded in self.keep_onloaded_values:
            return self.keep_onloaded_values[offloaded]

        # onload value
        onloaded = self.onload(offloaded)

        # when offloading is disabled, populate cache
        if self.offloading_disabled:
            self.keep_onloaded_values[offloaded] = onloaded

        return onloaded

    def __setitem__(self, key: Hashable, value: torch.Tensor | None):
        """
        Update the offloaded and onloaded values if the key exists, otherwise
        offload the value and add it to the cache.

        If called within the `disable_onloading` context, the value is assigned directly

        :param key: name of tensor
        :param value: tensor value to offload
        """
        # when onloading is disabled, parameters can be access and assigned directly
        if self.onloading_disabled:
            self.offloaded_values[key] = value
            return

        # if the key already exists, update with the new value
        offloaded = self.offloaded_values.get(key, None)
        if offloaded is not None and torch.is_same_size(offloaded, value):
            self.update_offload(offloaded, value)

            onloaded = self.keep_onloaded_values.get(offloaded, None)
            if onloaded is not None and onloaded is not offloaded:
                onloaded.copy_(value)

        # if the key does not exist (or the value is None), offload the new value
        else:
            self.offloaded_values[key] = self.offload(value)

    def __delitem__(self, key: Hashable):
        """
        Remove the offloaded tensor associated with `key`. Any references to its
        onloaded tensors held by this class are invalidated.

        :param key: name of tensor to invalidate
        """
        offloaded = self.offloaded_values[key]
        del self.offloaded_values[key]

        # remove strong ref
        if offloaded in self.keep_onloaded_values:
            del self.keep_onloaded_values[offloaded]

    def __contains__(self, key) -> bool:
        return key in self.offloaded_values

    def __iter__(self):
        return iter(self.offloaded_values)

    def __len__(self):
        return len(self.offloaded_values)

    @classmethod
    @contextlib.contextmanager
    def disable_offloading(cls):
        """
        Context to disable all offloading for offloaded modules which share this cache.
        After a weight has been fetched once, that onloaded value is cached and
        subsequent fetches will leverage the cache, reducing device movement
        """
        if not OffloadCache.offloading_disabled:
            restore_value = OffloadCache.offloading_disabled
            OffloadCache.offloading_disabled = True
            yield
            OffloadCache.offloading_disabled = restore_value
            OffloadCache.keep_onloaded_values.clear()
        else:
            yield

    @classmethod
    @contextlib.contextmanager
    def disable_onloading(cls):
        """
        Context to disable all onloading for offloaded modules which share this cache.
        This is mostly used for debugging purposes, and allows the caller to directly
        inspect offloaded tensors and directly assign offloaded tensors without copying
        """
        if not OffloadCache.onloading_disabled:
            restore_value = OffloadCache.onloading_disabled
            OffloadCache.onloading_disabled = True
            yield
            OffloadCache.onloading_disabled = restore_value
        else:
            yield



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import os

import pytest
import torch
import torch.distributed as dist
from compressed_tensors.offload import disable_onloading
from compressed_tensors.offload.cache.disk import DiskCache
from compressed_tensors.offload.cache.dist_disk import DistributedDiskCache
from safetensors import safe_open
from tests.test_offload.cache.helpers import (
    _test_delete,
    _test_disable_offloading,
    _test_disable_onloading,
    _test_garbage_collect,
    _test_offload,
    _test_onload,
    _test_onloading,
    _test_shared_attributes,
    _test_tensor_subclass,
)
from tests.test_offload.conftest import assert_tensor_equal, torchrun
from tests.testing_utils import requires_gpu


@pytest.fixture()
def onload_device():
    return torch.device("cuda")


@pytest.fixture()
def offload_device():
    return "disk"


@pytest.mark.unit
@requires_gpu(2)
@torchrun(world_size=2)
def test_delete(offload_device, onload_device, offload_cache):
    _test_delete(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu(2)
@torchrun(world_size=2)
def test_disable_offloading(offload_device, onload_device, offload_cache):
    _test_disable_offloading(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu(2)
@torchrun(world_size=2)
def test_disable_onloading(offload_device, onload_device, offload_cache):
    _test_disable_onloading(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu(2)
@torchrun(world_size=2)
def test_garbage_collect(offload_device, onload_device, offload_cache):
    _test_garbage_collect(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu(2)
@torchrun(world_size=2)
def test_offload(offload_device, onload_device, offload_cache):
    _test_offload(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu(2)
@torchrun(world_size=2)
def test_onload(offload_device, onload_device, offload_cache):
    _test_onload(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu(2)
@torchrun(world_size=2)
def test_onloading(offload_device, onload_device, offload_cache):
    _test_onloading(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu(2)
@torchrun(world_size=2)
def test_shared_attributes(offload_device, onload_device, offload_cache):
    _test_shared_attributes(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu(2)
@torchrun(world_size=2)
def test_tensor_subclass(offload_device, onload_device, offload_cache):
    _test_tensor_subclass(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu(2)
@torchrun(world_size=2)
def test_distributed_offload(onload_device, tmp_path):
    # Broadcast directory path from rank 0 to all ranks
    if dist.get_rank() == 0:
        offload_dir = tmp_path / "offload_dir"
        os.mkdir(offload_dir)
        broadcast_obj = [str(offload_dir)]
    else:
        broadcast_obj = [None]

    dist.broadcast_object_list(broadcast_obj, src=0)
    offload_dir = broadcast_obj[0]

    # Ensure directory creation completes before other ranks proceed
    dist.barrier()

    cache = DistributedDiskCache(onload_device, offload_dir=offload_dir)
    tensor = torch.zeros((5, 2))
    cache["tensor"] = tensor

    # check tensor construction
    assert torch.equal(cache["tensor"], tensor.to(onload_device))
    with disable_onloading():
        assert_tensor_equal(cache["tensor"], tensor.to("meta"))

    # update tensor
    tensor = torch.ones((5, 2))
    cache["tensor"] = tensor

    # check tensor construction
    assert torch.equal(cache["tensor"], tensor.to(onload_device))
    with disable_onloading():
        assert_tensor_equal(cache["tensor"], tensor.to("meta"))


@pytest.mark.unit
@requires_gpu(2)
@torchrun(world_size=2)
def test_distributed_files(tmp_path):
    # Broadcast directory path from rank 0 to all ranks
    if dist.get_rank() == 0:
        offload_dir = tmp_path / "offload_dir"
        os.mkdir(offload_dir)
        broadcast_obj = [str(offload_dir)]
    else:
        broadcast_obj = [None]

    dist.broadcast_object_list(broadcast_obj, src=0)
    offload_dir = broadcast_obj[0]

    # Ensure directory creation completes before other ranks proceed
    dist.barrier()

    # initial write, broadcasted to all ranks
    DiskCache.index = {}
    cache = DistributedDiskCache("cpu", offload_dir=offload_dir)
    tensor = torch.zeros(10)
    cache["weight"] = tensor

    assert len(DiskCache.index) == 1
    if dist.get_rank() == 0:  # only rank0 bc `tmp_path` is not shared between ranks
        files = os.listdir(offload_dir)
        assert len(files) == 1
        with safe_open(
            os.path.join(offload_dir, files[0]), framework="pt", device="cpu"
        ) as file:
            read_tensor = file.get_tensor("weight")
            assert_tensor_equal(read_tensor, tensor)

    # modify on one rank
    tensor = torch.ones(10)
    if dist.get_rank() == 0:
        cache["weight"] = tensor

    assert len(DiskCache.index) == 1
    if dist.get_rank() == 0:  # only rank0 bc `tmp_path` is not shared between ranks
        files = os.listdir(offload_dir)
        assert len(files) == 1
        with safe_open(
            os.path.join(offload_dir, files[0]), framework="pt", device="cpu"
        ) as file:
            read_tensor = file.get_tensor("weight")
            assert_tensor_equal(read_tensor, tensor)

    # delete
    del cache["weight"]
    assert len(DiskCache.index) == 0
    if dist.get_rank() == 0:  # only rank0 bc `tmp_path` is not shared between ranks
        files = os.listdir(offload_dir)
        assert len(files) == 0


@pytest.mark.unit
@requires_gpu(2)
@torchrun(world_size=2)
def test_distributed_async_update(tmp_path):
    """
    Test that different ranks can update different tensors asynchronously,
    and that values are correct after a barrier.
    """
    offload_dir = tmp_path / "offload_dir"
    if dist.get_rank() == 0:
        os.mkdir(offload_dir)

    # Ensure directory creation completes before other ranks proceed
    dist.barrier()

    onload_device = torch.device("cuda")
    cache = DistributedDiskCache(onload_device, offload_dir=str(offload_dir))

    # Initialize two tensors in the cache
    cache["tensor_0"] = torch.zeros(10, device=onload_device)
    cache["tensor_1"] = torch.zeros(10, device=onload_device)

    # Each rank updates a different tensor
    rank = dist.get_rank()
    if rank == 0:
        # Rank 0 updates tensor_0
        cache[f"tensor_{rank}"] = torch.ones(10, device=onload_device) * 1.0
    elif rank == 1:
        # Rank 1 updates tensor_1
        cache[f"tensor_{rank}"] = torch.ones(10, device=onload_device) * 2.0

    # Synchronize to ensure all updates are complete
    dist.barrier()

    # Verify that both tensors have the correct values on all ranks
    tensor_0 = cache["tensor_0"]
    tensor_1 = cache["tensor_1"]

    assert torch.allclose(tensor_0.cpu(), torch.ones(10) * 1.0)
    assert torch.allclose(tensor_1.cpu(), torch.ones(10) * 2.0)

    # Verify offloaded values are also correct
    with disable_onloading():
        offloaded_0 = cache["tensor_0"]
        offloaded_1 = cache["tensor_1"]
        assert_tensor_equal(offloaded_0, torch.ones(10) * 1.0, "disk")
        assert_tensor_equal(offloaded_1, torch.ones(10) * 2.0, "disk")



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import torch
import torch.distributed as dist
from compressed_tensors.offload.cache.disk import DiskCache
from compressed_tensors.offload.utils import send_tensors


class DistributedDiskCache(DiskCache):
    """
    Handles offloading and onloading tensors from/to disk. For more information, see
    `compressed_tensors.offload.cache.disk_cache::DiskCache`.
    """

    def offload(self, tensor: torch.Tensor | None) -> torch.Tensor | None:
        """
        Synchronously write tensor data to disk

        :param tensor: tensor on any device
        :return: meta tensor representing disk offloaded parameter
        """
        if tensor is None:
            return None

        if dist.get_rank() == 0:
            # write to disk
            offloaded = super().offload(tensor)
            broadcast_obj = [
                self.index[offloaded]["safetensors_file"],
                self.index[offloaded]["weight_name"],
                self.index[offloaded]["dtype"],
            ]
        else:
            offloaded = send_tensors(tensor, device="meta")
            broadcast_obj = [None, None, None]

        dist.broadcast_object_list(broadcast_obj, src=0)

        if dist.get_rank() != 0:
            self.index[offloaded] = {
                "safetensors_file": broadcast_obj[0],
                "weight_name": broadcast_obj[1],
                "dtype": broadcast_obj[2],
            }

        # wait for write to finish
        dist.barrier()
        return offloaded

    def __delitem__(self, key: str):
        """
        Remove the offload associated with `key`. If a new file was created to store
        updated tensor data, that new tensor data file is deleted.

        Any references to onloaded tensors held by this class are invalidated.

        :param key: name of tensor to invalidate
        """
        if dist.get_rank() == 0:
            super().__delitem__(key)
        else:
            offloaded = self.offloaded_values[key]
            del self.index[offloaded]
            super(DiskCache, self).__delitem__(key)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import torch
import torch.distributed as dist
from compressed_tensors.offload.cache.cpu import CPUCache
from compressed_tensors.offload.utils import send_tensors, to_empty


class DistributedCPUCache(CPUCache):
    """
    Handles offloading and onloading tensors from/to cpu memory shared across processes
    """

    def offload(self, tensor: torch.Tensor | None) -> torch.Tensor | None:
        """
        Synchronously create shared cpu memory for offload

        :param tensor: tensor on any device
        :return: cpu tensor whose data is located in shared memory
        """
        if tensor is None:
            return None

        # slight runtime cost for views
        tensor = tensor.contiguous()

        if dist.get_rank() == 0:
            # create shared memory cpu tensor
            tensor = super().offload(tensor).share_memory_()
            (handle, filename, nbytes) = tensor.untyped_storage()._share_filename_cpu_()
            broadcast_obj = [handle, filename, nbytes]
        else:
            broadcast_obj = [None, None, None]

        # receive shared memory file handle
        dist.broadcast_object_list(broadcast_obj, src=0)

        if dist.get_rank() != 0:
            # materialize meta tensor only if necessary
            if tensor.device.type == "meta":
                tensor = to_empty(tensor, device=self.offload_device)
            else:
                tensor = send_tensors(tensor, device=self.offload_device)

            # reconstruct tensor from shared memory file handle
            with torch.no_grad():
                tensor.set_(
                    torch.UntypedStorage._new_shared_filename_cpu(*broadcast_obj),
                    storage_offset=tensor.storage_offset(),
                    size=tensor.size(),
                    stride=tensor.stride(),
                )

        # ensure that rank 0 does not garbage collect before other ranks reconstruct
        dist.barrier()

        return tensor



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from unittest.mock import patch

import pytest
import torch
from compressed_tensors.offload.cache import CPUCache, OffloadCache
from compressed_tensors.offload.dispatch import (
    dispatch_model,
    get_device_memory,
    offload_model,
)
from compressed_tensors.offload.utils import module_size
from tests.testing_utils import requires_gpu
from transformers import AutoModelForCausalLM, AutoTokenizer


class Decoder(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.linear0 = torch.nn.Linear(5, 5)
        self.linear1 = torch.nn.Linear(5, 5)

    def forward(self, input):
        return self.linear1(self.linear0(input))


class Model(torch.nn.Module):
    _no_split_modules = ["Decoder"]

    def __init__(self):
        super().__init__()
        self.decoder0 = Decoder()
        self.decoder1 = Decoder()

    def forward(self, input):
        return self.decoder1(self.decoder0(input))


def assert_module_on_device(module: torch.nn.Module, device: torch.device | str):
    assert not isinstance(module._parameters, CPUCache)
    for name, param in module.named_parameters():
        assert torch.device(param.device) == torch.device(device), name


def assert_module_offloaded(
    module: torch.nn.Module,
    onload_device: torch.device | str,
    offload_device: torch.device | str,
    req_params: bool = False,
):
    for name, submodule in module.named_modules():
        if isinstance(submodule, torch.nn.ModuleList):
            continue
        if req_params and module_size(submodule)[0] <= 0:
            continue

        assert isinstance(submodule._parameters, OffloadCache), name
        assert torch.device(submodule._parameters.onload_device) == torch.device(
            onload_device
        )
        assert torch.device(submodule._parameters.offload_device) == torch.device(
            offload_device
        )


def has_memory_requirements(device_memory: dict[torch.device, int]):
    real_device_memory = get_device_memory()
    for key, req in device_memory.items():
        if key not in real_device_memory or real_device_memory[key] < req:
            return False

    return True


@pytest.mark.unit
@requires_gpu
def test_dispatch_one_device():
    model = Model()
    device_memory = {torch.device("cuda:0"): module_size(model)}
    if not has_memory_requirements(device_memory):
        pytest.skip("Cannot perform one device dispatch test, not enough device memory")

    dispatch_model(model, device_memory=device_memory, extra_memory=0)
    assert_module_on_device(model, "cuda:0")


@pytest.mark.unit
@requires_gpu
def test_dispatch_two_devices():
    model = Model()
    device_memory = {
        torch.device("cuda:0"): module_size(model.decoder0),
        torch.device("cuda:1"): module_size(model) - module_size(model.decoder0),
    }
    if not has_memory_requirements(device_memory):
        pytest.skip("Cannot perform split dispatch test: not enough devices or memory")

    # first decoder on first device, rest on second device
    dispatch_model(model, device_memory=device_memory, extra_memory=0)
    assert_module_on_device(model.decoder0, "cuda:0")
    assert_module_on_device(model.decoder1, "cuda:1")


@pytest.mark.unit
@requires_gpu
def test_dispatch_no_split():
    model = Model()
    device_memory = {
        torch.device("cuda:0"): module_size(model.decoder0.linear0),
        torch.device("cuda:1"): module_size(model),
    }
    if not has_memory_requirements(device_memory):
        pytest.skip("Cannot perform split dispatch test: not enough devices or mem")

    # first device is skipped: all ends up on second device
    dispatch_model(model, device_memory=device_memory, extra_memory=0)
    assert_module_on_device(model, "cuda:1")


@pytest.mark.unit
@requires_gpu
def test_dispatch_split():
    model = Model()
    first_linear = model.decoder0.linear0
    device_memory = {
        torch.device("cuda:0"): module_size(first_linear),
        torch.device("cuda:1"): module_size(model) - module_size(first_linear),
    }
    if not has_memory_requirements(device_memory):
        pytest.skip("Cannot perform split dispatch test: not enough devices or memory")

    # first linear on first device, rest on second device
    dispatch_model(
        model, device_memory=device_memory, no_split_modules=tuple(), extra_memory=0
    )
    assert_module_on_device(model.decoder0.linear0, "cuda:0")
    assert_module_on_device(model.decoder0.linear1, "cuda:1")
    assert_module_on_device(model.decoder1, "cuda:1")


@pytest.mark.unit
@requires_gpu
def test_dispatch_offloaded():
    model = Model()
    device_memory = {
        torch.device("cuda:0"): (
            module_size(model.decoder0.linear0) + module_size(model.decoder1)
        ),
    }
    if not has_memory_requirements(device_memory):
        pytest.skip("Cannot perform split dispatch test: not enough devices or mem")

    with patch("compressed_tensors.offload.dispatch.get_module_sizes") as mock_sizes:
        # first two linears are disjoint, but not enough memory to fit decoder1
        mock_sizes.return_value = [
            (model.decoder0.linear0, module_size(model.decoder0.linear0)),
            (model.decoder0.linear1, module_size(model.decoder0.linear1)),
            (model.decoder1, module_size(model.decoder1)),
        ]

        # first linear stays onloaded
        # second linear is popped off to fit offloaded decoder1
        dispatch_model(
            model, device_memory=device_memory, no_split_modules=tuple(), extra_memory=0
        )
        assert_module_on_device(model.decoder0.linear0, "cuda:0")
        assert_module_offloaded(model.decoder0.linear1, "cuda:0", "cpu")
        assert_module_offloaded(model.decoder1, "cuda:0", "cpu")


@pytest.mark.integration
@requires_gpu
@pytest.mark.parametrize("model_id", ["nm-testing/tinysmokellama-3.2"])
@torch.inference_mode()
def test_offload_and_dispatch_model(model_id):
    model = AutoModelForCausalLM.from_pretrained(model_id).eval()
    tokenizer = AutoTokenizer.from_pretrained(model_id)

    tied_tensors_size = model.lm_head.weight.nbytes
    device_memory = {torch.device("cuda:0"): module_size(model) + tied_tensors_size}
    if not has_memory_requirements(device_memory):
        pytest.skip("Cannot perform split dispatch test: not enough devices or mem")

    model.to("cuda:0")
    sample = tokenizer("Hello my name is", return_tensors="pt")
    sample = {k: v.to("cuda:0") for k, v in sample.items()}
    true_logits = model(**sample).logits

    # offload entire model
    model.to("cpu")
    model = offload_model(model, "cuda:0")
    offloaded_logits = model(**sample).logits
    for module in model.modules():
        assert_module_offloaded(module, "cuda:0", torch.device("cpu"))
    assert torch.allclose(offloaded_logits, true_logits)

    # dispatch model and fits
    model = dispatch_model(model, device_memory=device_memory, extra_memory=0)
    dispatched_logits = model(**sample).logits
    for module in model.modules():
        assert_module_on_device(module, "cuda:0")
    assert torch.allclose(dispatched_logits, true_logits)

    # dispatch model with offload
    device_memory[torch.device("cuda:0")] = device_memory[torch.device("cuda:0")] // 2
    model = dispatch_model(model, device_memory=device_memory, extra_memory=0)
    dispatched_logits = model(**sample).logits
    assert torch.allclose(dispatched_logits, true_logits)


@pytest.mark.unit
def test_get_device_memory_cpu_fallback():
    with patch("compressed_tensors.offload.dispatch.torch.cuda") as mock_cuda:
        mock_cuda.is_available.return_value = False
        device_memory = get_device_memory()

    assert len(device_memory) == 1
    assert torch.device("cpu") in device_memory
    assert device_memory[torch.device("cpu")] > 0


@pytest.mark.unit
def test_dispatch_cpu_only():
    model = Model()
    cpu_memory = module_size(model) * 2
    device_memory = {torch.device("cpu"): cpu_memory}

    dispatch_model(model, device_memory=device_memory, extra_memory=0)
    assert_module_on_device(model, "cpu")


@pytest.mark.unit
def test_dispatch_cpu_only_via_fallback():
    model = Model()

    with patch("compressed_tensors.offload.dispatch.torch.cuda") as mock_cuda:
        mock_cuda.is_available.return_value = False
        dispatch_model(model, extra_memory=0)

    assert_module_on_device(model, "cpu")



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from collections import defaultdict
from enum import Enum
from typing import Annotated, Any

from compressed_tensors.config import CompressionFormat
from compressed_tensors.quantization.quant_args import DynamicType, QuantizationArgs
from compressed_tensors.quantization.quant_scheme import (
    QuantizationScheme,
    preset_name_to_scheme,
)
from compressed_tensors.quantization.utils import is_module_quantized, module_type
from pydantic import BaseModel, ConfigDict, Field
from torch.nn import Module


__all__ = [
    "QuantizationStatus",
    "QuantizationConfig",
    "LIFECYCLE_ORDER",
    "DEFAULT_QUANTIZATION_METHOD",
    "DEFAULT_QUANTIZATION_FORMAT",
]


class QuantizationStatus(str, Enum):
    """
    Enum storing the different states a quantized layer can be in

    - Initialized: Quantization parameters are initialized to empty values
    - Calibration: Quantization parameters are being calibrated, observers are attached
    - Frozen: Quantization parameters are fully calibrated, observers are removed
    - Compressed: All parameters are quantized to target dtype
    - Decompressed: Parameters are converted back into frozen state,
        additionally, weight qdq is skipped during forward passes for better performance
    """

    INITIALIZED = "initialized"
    CALIBRATION = "calibration"
    FROZEN = "frozen"
    COMPRESSED = "compressed"
    DECOMPRESSED = "decompressed"

    @classmethod
    def lifecycle_order(cls) -> list["QuantizationStatus"]:
        """
        :return: list of correct quantization lifecycle order
        """
        return

    def __ge__(self, other):
        if other is None:
            return True
        if not isinstance(other, self.__class__):
            raise NotImplementedError
        return LIFECYCLE_ORDER.index(self) >= LIFECYCLE_ORDER.index(other)

    def __gt__(self, other):
        if other is None:
            return True
        if not isinstance(other, self.__class__):
            raise NotImplementedError
        return LIFECYCLE_ORDER.index(self) > LIFECYCLE_ORDER.index(other)

    def __lt__(self, other):
        if other is None:
            return False
        if not isinstance(other, self.__class__):
            raise NotImplementedError
        return LIFECYCLE_ORDER.index(self) < LIFECYCLE_ORDER.index(other)

    def __le__(self, other):
        if other is None:
            return False
        if not isinstance(other, self.__class__):
            raise NotImplementedError
        return LIFECYCLE_ORDER.index(self) <= LIFECYCLE_ORDER.index(other)


LIFECYCLE_ORDER = [
    QuantizationStatus.INITIALIZED,
    QuantizationStatus.CALIBRATION,
    QuantizationStatus.FROZEN,
    QuantizationStatus.COMPRESSED,
    QuantizationStatus.DECOMPRESSED,
]

DEFAULT_QUANTIZATION_METHOD = "compressed-tensors"
DEFAULT_QUANTIZATION_FORMAT = "fakequant"


class QuantizationConfig(BaseModel):
    """
    Full configuration specifying how a model is quantized. Each quantized layer is
    mapped to a QuantizationScheme in config_groups.

    :param config_groups: dict of QuantizationSchemes specifying the quantization
    settings for each quantized layer. A group could also be a reference to
    a predefined scheme name, mapped to a list of its target layers/classes
    :param quant_method: a constant used to differentiate compressed-tensors
    quantization from other quantization configs
    :param format: specifies how the quantized model is stored on disk
    :quantization_status: specifies the current status of all quantized layers. It is
        assumed all layers are in the same state.
    :param kv_cache_scheme: optional QuantizationArgs, that specify the
        quantization of the kv cache. If None, kv cache is not quantized.
        When applying kv cache quantization to transformer AutoModelForCausalLM,
        the kv_cache_scheme gets converted into a QuantizationScheme that:
            - targets the `q_proj` and `k_proj` modules of the model. The outputs
              of those modules are the keys and values that might be cached
            - quantizes the outputs of the aformentioned layers, so that
              keys and values are compressed before storing them in the cache
        There is an explicit assumption that the model contains modules with
        `k_proj` and `v_proj` in their names. If this is not the case
        and kv_cache_scheme != None, the quantization of kv cache will fail
    :global_compression_ratio: optional informational config to report the model
        compression ratio acheived by the quantization config
    :ignore: optional list of layers to ignore from config_groups. Layers in this list
        are not quantized even if they match up with a target in config_groups
    """

    config_groups: dict[str, QuantizationScheme | list[str]]
    quant_method: str = DEFAULT_QUANTIZATION_METHOD
    kv_cache_scheme: QuantizationArgs | None = None
    format: str = DEFAULT_QUANTIZATION_FORMAT
    quantization_status: QuantizationStatus = QuantizationStatus.INITIALIZED
    global_compression_ratio: float | None = None
    ignore: list[str] | None = Field(default_factory=list)
    # `run_compressed` is a dummy, unused arg for backwards compatibility
    # see: https://github.com/huggingface/transformers/pull/39324
    run_compressed: Annotated[Any, Field(exclude=True)] = None

    def model_post_init(self, __context):
        """
        updates any quantization schemes defined as presets to be fully loaded
        schemes
        """
        for group_name, targets_or_scheme in self.config_groups.items():
            if isinstance(targets_or_scheme, QuantizationScheme):
                continue  # scheme already defined
            self.config_groups[group_name] = preset_name_to_scheme(
                name=group_name,
                targets=targets_or_scheme,
            )

    def to_dict(self):
        # for compatibility with HFQuantizer
        return self.model_dump()

    @staticmethod
    def from_pretrained(
        model: Module, format: str | list | None = None
    ) -> "QuantizationConfig | None":
        """
        Converts a model into its associated QuantizationConfig based on the
        QuantizationScheme attached to each quantized module

        :param model: model to calculate quantization scheme of
        :return: filled out QuantizationScheme for the input model
        """
        from compressed_tensors.modeling import IMPL_ATTR
        from compressed_tensors.quantization.lifecycle.initialize import (
            is_attention_module,
        )

        # set of all quantization schemes
        # TODO: make quant config/scheme/args frozen/hashable and use a set
        quantization_schemes: list[QuantizationScheme] = list()

        # use any status from modules (in practice, use the last module)
        model_status = None

        # set of all quantized types
        # this is later used to create the ignore list
        quantization_type_names: set[str] = set()

        # maps types to names which are not quantized
        # this is later used to create the ignore list
        ignore: dict[str, list[str]] = defaultdict(list)

        # this keeps track of any kvcache schemes
        kv_cache_scheme: QuantizationArgs | None = None

        for name, submodule in model.named_modules():
            layer_type: str = module_type(submodule)

            # add config group if quantized non-attention or attention quant
            has_config_group = is_module_quantized(submodule) and (
                not is_attention_module(submodule) or hasattr(submodule, IMPL_ATTR)
            )
            # only add kvcache if quant attention (which always implies kvcache)
            has_kv_cache = is_module_quantized(submodule) and is_attention_module(
                submodule
            )

            if has_config_group:
                # add to running set of schemes/layer_type_names
                model_status = getattr(submodule, "quantization_status", model_status)
                quantization_type_names.add(layer_type)
                if submodule.quantization_scheme not in quantization_schemes:
                    quantization_schemes.append(submodule.quantization_scheme)

            if has_kv_cache:
                model_status = getattr(submodule, "quantization_status", model_status)
                kv_cache_scheme = submodule.quantization_scheme.input_activations

            if not has_config_group:
                # add non-quantized layers to the ignore list
                if layer_type not in ignore:
                    ignore[layer_type] = []
                ignore[layer_type].append(name)

        if (
            len(quantization_schemes) == 0 and kv_cache_scheme is None
        ):  # No quantized layers
            return None

        # create ignore list, only include layers whose class has ever been targeted
        consolidated_ignore = []
        for layer_type, ignore_names in ignore.items():
            if layer_type in quantization_type_names:
                # specific layers of a quantized type are ignored
                consolidated_ignore += ignore_names
            # else we leave it off the ignore list, doesn't fall under any of the
            # existing quantization schemes so it won't be quantized

        # create config groups from all unique schemes
        config_groups = {}
        for idx, scheme in enumerate(quantization_schemes):
            group_name = "group_" + str(idx)
            config_groups[group_name] = scheme

        # infer format
        if format is None:
            if model_status == QuantizationStatus.COMPRESSED:
                format = CompressionFormat.int_quantized.value
            else:
                format = CompressionFormat.dense.value
        elif isinstance(format, list):
            format = (
                CompressionFormat.mixed_precision.value
                if len(format) > 1
                else format[0]
            )

        return QuantizationConfig(
            config_groups=config_groups,
            quantization_status=model_status,
            kv_cache_scheme=kv_cache_scheme,
            global_compression_ratio=None,
            format=format,
            ignore=consolidated_ignore,
        )

    def requires_calibration_data(self):
        if self.kv_cache_scheme is not None:
            return True

        for _, scheme in self.config_groups.items():
            if scheme.input_activations is not None:
                if scheme.input_activations.dynamic in (False, DynamicType.LOCAL):
                    return True
            if scheme.output_activations is not None:
                if not scheme.output_activations.dynamic:
                    return True

        return False

    # TODO set `extra="forbid"` when upstream transformers is compatible
    model_config = ConfigDict(extra="ignore")



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from collections.abc import Generator
from pathlib import Path

import torch
from compressed_tensors.config import CompressionFormat
from compressed_tensors.utils import deprecated
from torch import Tensor


__all__ = [
    "load_compressed",
    "save_compressed",
    "save_compressed_model",
]


@deprecated("LLM Compressor's `model_free_ptq` pathway")
def save_compressed(
    tensors: dict[str, Tensor],
    save_path: str | Path,
    compression_format: CompressionFormat | None = None,
) -> None:
    raise NotImplementedError()


@deprecated("LLM Compressor's `model_free_ptq` pathway")
def load_compressed(
    compressed_tensors: str | Path,
    compression_config: object = None,
    device: str | None = "cpu",
) -> Generator[tuple[str, Tensor], None, None]:
    raise NotImplementedError()


@deprecated("LLM Compressor's `model_free_ptq` pathway")
def save_compressed_model(
    model: torch.nn.Module,
    filename: str,
    compression_format: CompressionFormat | None = None,
    force_contiguous: bool = True,
) -> None:
    raise NotImplementedError()



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import json
import os

from compressed_tensors import __version__ as ct_version
from compressed_tensors.base import COMPRESSION_VERSION_NAME, QUANTIZATION_CONFIG_NAME
from compressed_tensors.entrypoints.convert import (
    Converter,
    find_config_path,
    find_safetensors_index_path,
)
from loguru import logger


__all__ = ["update_config", "update_safetensors_index"]


def update_config(
    save_directory: str | os.PathLike,
    converter: Converter,
):
    """
    Update Quantization config for model stub in save_directory,
    based on the converter that was used.
    Quantization config is considered stale and re-written entirely.
    """
    quant_config = converter.create_config()

    quant_config_data = quant_config.model_dump()
    quant_config_data[COMPRESSION_VERSION_NAME] = ct_version

    # write results to config.json or params.json file
    config_file_path = find_config_path(save_directory)
    if config_file_path is not None:
        with open(config_file_path, "r") as file:
            config_data = json.load(file)

        config_data[QUANTIZATION_CONFIG_NAME] = quant_config_data

        with open(config_file_path, "w") as file:
            json.dump(config_data, file, indent=2, sort_keys=True)

    else:
        logger.warning(
            f"Could not find config file in {save_directory}. Please add to config "
            f"{json.dumps(quant_config_data, indent=2, sort_keys=True)}"
        )


def update_safetensors_index(
    save_directory: str | os.PathLike,
    total_size: int,
    weight_map: dict[str, str],
):
    file_path = find_safetensors_index_path(save_directory)
    if file_path is None:
        return

    with open(file_path, "w") as file:
        json.dump(
            {
                "metadata": {
                    "total_size": total_size,
                },
                "weight_map": weight_map,
            },
            file,
            indent=2,
            sort_keys=True,
        )



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

# flake8: noqa
# isort: skip_file

from .quant_args import *
from .quant_config import *
from .quant_metadata import *
from .quant_scheme import *
from .lifecycle import *



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import math

import torch
from compressed_tensors.quantization.quant_args import (
    BFLOAT16_DATA,
    FLOAT32_DATA,
    FP4_E2M1_DATA,
    FP8_E4M3_DATA,
    QuantizationArgs,
    QuantizationType,
)


__all__ = [
    "maybe_convert_from_mx_exp",
    "generate_mx_scales",
    "round_to_power_2",
    "should_generate_mx_scales",
]

# Reference: https://github.com/vllm-project/vllm/blob/main/tests/quantization/reference_mxfp4.py # noqa: E501

# The exponent offset maps the group max into the quantized type's
# representable range.  It equals floor(log2(type_max)):
#   FP4 E2M1  max=6.0   -> floor(log2(6))   = 2
#   FP8 E4M3  max=448.0 -> floor(log2(448)) = 8
_MX_ELEM_OFFSET = {
    4: int(math.floor(math.log2(FP4_E2M1_DATA.max))),  # 2
    8: int(math.floor(math.log2(FP8_E4M3_DATA.max))),  # 8
}


def should_generate_mx_scales(args: QuantizationArgs):
    return (
        args.num_bits in (4, 8)
        and args.type == QuantizationType.FLOAT.value
        and args.group_size == 32
        and args.scale_dtype == torch.uint8
    )


def maybe_convert_from_mx_exp(
    args: QuantizationArgs, scale: torch.Tensor
) -> torch.Tensor:
    """
    Conditionally converts MX (MXFP4/MXFP8) scales from their E8M0 exponent
    format to float scales.

    If the quantization arguments indicate an MX format, the input `scale`
    is treated as E8M0 uint8 exponents and converted to float power-of-2
    scales. Otherwise, the input `scale` tensor is returned unchanged.

    :param args: quantization args to check for MX format
    :param scale: tensor of scale values (uint8 exponents for MX, or float)
    :return: float scale tensor, or original scale if not MX format
    """
    original_dtype = scale.dtype
    if should_generate_mx_scales(args):
        scale_exp = scale.to(torch.int32) - 127
        scale = 2.00 ** (scale_exp.to(torch.float))
        return scale.to(original_dtype)
    return scale


def round_to_power_2(x: torch.Tensor) -> torch.Tensor:
    """
    Round values to the closest power of 2.
    This is done by masking the values with BFLOAT16_SIGN_EXPONENT_MASK
    which essentially removes the mantissa and keeps the exponent.
    i.e the closest power of 2 for the input_value.

    E.g:
        0.0825 = 1.32 (mantissa) x 2**-4 (exponent)
        0.0825 ==> -4 (exponent) + 127 = 123 = 01111011 (8 bits for bfloat16)
        0.0825 ==> 0.32 (mantissa) = 0101001 (7 bits for bfloat16)
        0.0825 == 0b01111011_0101001 (bfloat16)
        0b01111011_0101001 & 111111111_0000000 == 0b01111011_0000000
        Keep the exponent + sign bit to give you the closest power of 2, 0.0625

    :param x: tensor to round to closest power of 2
    """
    scale_dtype = x.dtype
    if scale_dtype is torch.bfloat16:
        int_dtype = torch.uint16
        mantissa = BFLOAT16_DATA.mantissa
        exponent = BFLOAT16_DATA.exponent
    else:
        assert scale_dtype is torch.float32
        int_dtype = torch.uint32
        mantissa = FLOAT32_DATA.mantissa
        exponent = FLOAT32_DATA.exponent

    x = x.view(int_dtype).to(torch.int32)

    # Find closest power of 2
    VAL_TO_ADD = 1 << (mantissa - FP4_E2M1_DATA.mantissa - 1)
    # Add value to push the value to the next exponent
    SIGN_EXPONENT_MASK = ((1 << (exponent + 1)) - 1) << mantissa
    # mask to only keep exponent - we conservatively round down
    # to better represent smaller numbers / prevent overflow
    block_max_uint = torch.bitwise_and(x + VAL_TO_ADD, SIGN_EXPONENT_MASK)
    if scale_dtype is torch.bfloat16:
        return block_max_uint.to(int_dtype).view(scale_dtype)
    return block_max_uint.view(scale_dtype)


def generate_mx_scales(x: torch.Tensor, num_bits: int = 4) -> torch.Tensor:
    """
    Generate MX scales (for MXFP4 and MXFP8). The scales require the
    following steps:
    1. Round to the closest power of 2
    2. Subtract the element-format offset so that the largest group
       values map into the quantized type's representable range
    3. Convert to biased E8M0 exponent (bias 127)

    Called when calculating qparams using observers.

    :param x: tensor of per-group max absolute values
    :param num_bits: quantized element width (4 for MXFP4, 8 for MXFP8)
    :returns scales as E8M0 exponents (uint8 after rounding)
    """
    offset = _MX_ELEM_OFFSET[num_bits]
    # Round to closest power of 2
    scale_power_2 = round_to_power_2(x)
    return 127 + torch.floor(torch.log2(scale_power_2)) - offset



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from typing import TYPE_CHECKING, Literal, Optional

import torch
from compressed_tensors.offload.cache.base import OffloadCache
from compressed_tensors.offload.utils import send_tensors


if TYPE_CHECKING:
    from torch._prims_common import DeviceLikeType


class DeviceCache(OffloadCache):
    """
    Handles offloading and onloading tensors from/to device memory. Onloading
    tensors is typically a no-op (except onload device has been modified).
    """

    def __init__(
        self,
        onload_device: "DeviceLikeType",
        offload_device: Optional["DeviceLikeType | Literal['disk']"] = None,
    ):
        super().__init__(onload_device, offload_device=offload_device)
        if offload_device is not None:
            self.offload_device = torch.device(offload_device)
        else:
            self.offload_device = self.onload_device

    def onload(self, offloaded: torch.Tensor | None) -> torch.Tensor | None:
        """
        Typically a no op, except when onload device has been modified

        :param key: device tensor to onload
        :return: device tensor
        """
        # move because onload_device might be modified after init
        return send_tensors(offloaded, device=self.onload_device, copy=False)

    def offload(self, tensor: torch.Tensor | None) -> torch.Tensor | None:
        """
        Offload a tensor to the device

        :param value: tensor on any device
        :return: tensor on device
        """
        return send_tensors(tensor, device=self.offload_device, copy=False)

    @torch.no_grad()
    def update_offload(self, offloaded: torch.Tensor, data: torch.Tensor | None):
        """
        Update the device value with new data

        :param offloaded: device tensor to update
        :param data: new data to copy from
        """
        if data is not None:
            offloaded.copy_(data)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from typing import Iterable

import torch
from compressed_tensors.config import CompressionFormat
from compressed_tensors.entrypoints.convert.converters import Converter
from compressed_tensors.quantization import (
    QuantizationArgs,
    QuantizationConfig,
    QuantizationScheme,
    QuantizationStatus,
)
from compressed_tensors.quantization.quant_scheme import NVFP4
from compressed_tensors.utils.match import match_quantizable_tensors


class ModelOptNvfp4Converter(Converter):
    """
    Convert params from modelopt NVFP4 to CT NVFP4 convention,
    and optionally the kv_cache_scheme
    """

    def __init__(
        self,
        ignore: Iterable[str] = tuple(),
        targets: Iterable[str] = tuple(),
        kv_cache_scheme: QuantizationArgs | None = None,
    ):
        self.ignore = ignore
        self.targets = targets
        self.kv_cache_scheme = kv_cache_scheme

    def process(self, tensors: dict[str, torch.Tensor]):
        """
        Map the modelopt NVFP4 tensors to the appropriate compressed-tensors
        NVFP4 format.
        Some tensors require rename, some require inversion
        - 1 / input_scale -> input_global_scale
        - weight -> weight_packed
        - 1 / weight_scale_2 -> weight_global_scale
        """
        for module_name, name in match_quantizable_tensors(
            tensors, self.ignore, self.targets, allow_nonquantizable=True
        ):
            param_name = name.rsplit(".", 1)[-1]

            match param_name:
                # input_scale -> input_global_scale F32
                case "input_scale":
                    # convert modelopt input_scale x -> 1/x
                    # https://github.com/vllm-project/vllm/blob/v0.13.0/vllm/model_executor/layers/quantization/modelopt.py#L1070-L1073
                    # https://github.com/vllm-project/vllm/blob/v0.13.0/vllm/model_executor/layers/quantization/modelopt.py#L1134
                    # https://github.com/vllm-project/vllm/blob/v0.13.0/vllm/model_executor/layers/quantization/compressed_tensors/schemes/compressed_tensors_w4a4_nvfp4.py#L190
                    tensors[f"{module_name}.input_global_scale"] = 1 / tensors[name]
                    del tensors[name]
                # weight -> weight_packed U8
                case "weight":
                    tensors[f"{module_name}.weight_packed"] = tensors[name]
                    del tensors[name]
                # weight_scale -> weight_scale F8_E4M3
                case "weight_scale":
                    pass
                # weight_scale_2 -> weight_global_scale F32
                case "weight_scale_2":
                    # convert modelopt weight_scale_2 x -> 1/x
                    # https://github.com/vllm-project/vllm/blob/v0.13.0/vllm/model_executor/layers/quantization/modelopt.py#L1066-L1068
                    # https://github.com/vllm-project/vllm/blob/v0.13.0/vllm/model_executor/layers/quantization/compressed_tensors/schemes/compressed_tensors_w4a4_nvfp4.py#L163-L166
                    tensors[f"{module_name}.weight_global_scale"] = 1 / tensors[name]
                    del tensors[name]
                case "k_scale" | "v_scale":
                    # convert kv cache scales to appropriate dtype
                    # often F32 in modelopt, defaults BF16 in compressed-tensors
                    tensors[name] = tensors[name].to(
                        self.kv_cache_scheme.scale_dtype or torch.bfloat16
                    )

    def validate(self, tensors: dict[str, torch.Tensor]):
        """
        Ensure all tensor names of targeted layers are expected and no
        untargeted layers have unexpected tensor names
        """
        allowed_names = ["input_scale", "weight", "weight_scale", "weight_scale_2"]
        if self.kv_cache_scheme is not None:
            allowed_names += ["k_scale", "v_scale"]

        targeted_names = [
            name
            for _, name in match_quantizable_tensors(
                tensors, self.ignore, self.targets, allow_nonquantizable=True
            )
        ]
        for name in targeted_names:
            param_name = name.rsplit(".", 1)[-1]

            if param_name not in allowed_names:
                raise ValueError(f"Hit unexpected targeted tensor {name}")

        disallowed_names = ["input_scale", "weight_scale", "weight_scale_2"]
        untargeted_names = [
            name for name in tensors.keys() if name not in targeted_names
        ]
        for name in untargeted_names:
            param_name = name.rsplit(".", 1)[-1]

            if param_name in disallowed_names:
                raise ValueError(f"Hit unexpected non-targeted tensor {name}")

    def create_config(self) -> QuantizationConfig:
        return QuantizationConfig(
            config_groups={
                "config_group_0": QuantizationScheme(
                    **NVFP4,
                    targets=self.targets,
                    format=CompressionFormat.nvfp4_pack_quantized.value,
                )
            },
            ignore=self.ignore,
            kv_cache_scheme=self.kv_cache_scheme,
            format=CompressionFormat.nvfp4_pack_quantized.value,
            quantization_status=QuantizationStatus.COMPRESSED.value,
        )



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import math
from collections import OrderedDict

import pytest
import torch
from compressed_tensors import FloatQuantizationCompressor
from compressed_tensors.quantization import (
    QuantizationArgs,
    QuantizationConfig,
    QuantizationScheme,
    QuantizationStatus,
    QuantizationStrategy,
    apply_quantization_config,
)
from compressed_tensors.quantization.lifecycle.forward import fake_quantize
from torch.nn.modules import Linear, Sequential


def make_dummy_g_idx(columns: int, group_size: int) -> torch.Tensor:
    perm = torch.randperm(columns)
    return torch.tensor([index // group_size for index in range(columns)])[perm]


@pytest.mark.parametrize(
    "strategy,group_size,sc,zp",
    [
        [QuantizationStrategy.TENSOR, None, torch.tensor(0.01), torch.tensor(0)],
        [
            QuantizationStrategy.GROUP,
            128,
            torch.rand((512, 8)) * 0.01,
            torch.zeros((512, 8), dtype=torch.int8),
        ],
        [
            QuantizationStrategy.CHANNEL,
            None,
            torch.rand((512, 1)) * 0.01,
            torch.zeros((512, 1), dtype=torch.int8),
        ],
    ],
)
def test_quant_format(strategy, group_size, sc, zp):
    module_sd = {
        "weight": torch.rand((512, 1024)),
        "weight_scale": sc.to(torch.float32),
        "weight_zero_point": zp.to(torch.float32),
    }
    if group_size is not None:
        module_sd["weight_g_idx"] = make_dummy_g_idx(1024, group_size)

    scheme = QuantizationScheme(
        targets=["Linear"],
        weights=QuantizationArgs(
            strategy=strategy, type="float", group_size=group_size
        ),
    )

    compressed = FloatQuantizationCompressor.compress(module_sd, scheme=scheme)

    # symmetric → zero_point is dropped
    assert "weight_zero_point" not in compressed
    assert compressed["weight_scale"].dtype == torch.float32
    assert torch.equal(compressed["weight_scale"], module_sd["weight_scale"])
    if group_size is not None:
        assert torch.equal(compressed["weight_g_idx"], module_sd["weight_g_idx"])


@pytest.mark.parametrize(
    "strategy,group_size",
    [
        [QuantizationStrategy.TENSOR, None],
        [QuantizationStrategy.CHANNEL, None],
    ],
)
def test_compress_decompress_match(
    mock_per_group_calibration,
    mock_per_channel_calibration,
    strategy,
    group_size,
):
    model = Sequential(OrderedDict([("dummy", Linear(512, 1024, bias=None))]))
    quant_config = QuantizationConfig(
        config_groups={
            "group_1": QuantizationScheme(
                targets=["Linear"],
                weights=QuantizationArgs(
                    strategy=strategy, type="float", group_size=group_size
                ),
            )
        },
        ignore=["lm_head"],
    )
    apply_quantization_config(model, quant_config)
    model.dummy.quantization_status = QuantizationStatus.CALIBRATION

    if strategy == QuantizationStrategy.GROUP:
        mock_per_group_calibration(
            model.dummy, base_name="weight", value=model.dummy.weight, group_size=128
        )
    if strategy == QuantizationStrategy.CHANNEL:
        mock_per_channel_calibration(
            model.dummy, base_name="weight", value=model.dummy.weight
        )

    scheme = quant_config.config_groups["group_1"]

    # Build per-module state dict
    module_sd = {
        name: param.data.clone() for name, param in model.dummy.named_parameters()
    }

    compressed = FloatQuantizationCompressor.compress(module_sd, scheme=scheme)
    decompressed = FloatQuantizationCompressor.decompress(compressed, scheme=scheme)

    fake_quant_dummy = fake_quantize(
        model.dummy.weight,
        scale=model.dummy.weight_scale,
        zero_point=model.dummy.weight_zero_point,
        args=scheme.weights,
    )
    assert torch.equal(fake_quant_dummy, decompressed["weight"])


@pytest.mark.parametrize(
    "rows,cols,block_height,block_width",
    [
        (10944, 2048, 128, 128),
        (2048, 10944, 128, 128),
        (256, 256, 128, 128),
        (300, 400, 128, 128),
        (256, 300, 128, 128),
        (300, 256, 128, 128),
    ],
)
def test_block_quant_compression_padding(rows, cols, block_height, block_width):
    """
    Block quantization compresses weights with non-divisible dimensions without
    changing the original shape.
    """
    block_structure = [block_height, block_width]
    num_rb = math.ceil(rows / block_height)
    num_cb = math.ceil(cols / block_width)

    module_sd = {
        "weight": torch.rand((rows, cols)),
        "weight_scale": torch.rand((num_rb, num_cb)) * 0.01 + 0.001,
        "weight_zero_point": torch.zeros((num_rb, num_cb)),
    }

    scheme = QuantizationScheme(
        targets=["Linear"],
        weights=QuantizationArgs(
            strategy=QuantizationStrategy.BLOCK,
            type="float",
            block_structure=block_structure,
        ),
    )

    compressed = FloatQuantizationCompressor.compress(module_sd, scheme=scheme)

    # Compressed weight should retain the original shape
    assert compressed["weight"].shape == (
        rows,
        cols,
    ), "Compressed weight shape should match original"



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from itertools import chain

import torch
from compressed_tensors.utils.type import TensorStateDict


__all__ = ["get_direct_state_dict", "replace_direct_state_dict"]


def get_direct_state_dict(module: torch.nn.Module) -> TensorStateDict:
    """
    Extract a state dict directly from a module's parameters and buffers.

    Returns tensor data (unwrapped from Parameter/Buffer wrappers) for all
    parameters and buffers in the module. Does not recurse into child modules.

    :param module: the module to extract state from
    :return: dict mapping parameter/buffer names to their tensor data
    """
    return {
        name: (
            tensor.data
            if isinstance(tensor, (torch.nn.Parameter, torch.nn.Buffer))
            else tensor
        )
        for name, tensor in chain(module._parameters.items(), module._buffers.items())
    }


def replace_direct_state_dict(module: torch.nn.Module, new_state_dict: TensorStateDict):
    """
    Replace a module's parameters and buffers with a new state dict.

    Removes parameters/buffers that exist in the old state but not the new state,
    and adds/updates parameters from the new state dict. All new tensors are
    added as non-trainable parameters (not buffers). Skips unchanged values
    for efficiency.

    :param module: the module to update
    :param new_state_dict: dict of new parameter/buffer values
    """
    old_state_dict = get_direct_state_dict(module)

    for name, old_value in old_state_dict.items():
        # remove attributes that don't exist in the new state
        if name not in new_state_dict:
            delattr(module, name)

    for name, new_value in new_state_dict.items():
        # skip unchanged values
        if name not in old_state_dict or old_state_dict[name] is not new_value:
            # overwrite (not update) if param already existed
            if hasattr(module, name):
                delattr(module, name)

            # treat all new tensors as parameters (not buffers)
            setattr(module, name, torch.nn.Parameter(new_value, requires_grad=False))



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import torch
from compressed_tensors import TRANSFORM_CONFIG_NAME
from compressed_tensors.transform import TransformConfig, TransformFactory


__all__ = ["apply_transform_config"]


def apply_transform_config(model: torch.nn.Module, config: TransformConfig):
    """
    Apply a transform config to a model. Weight transforms are fused into weights, while
    activation transforms are attached as submodules and trigger via pytorch hooks

    :param model: model to apply config to
    :param config: transform config to apply
    """
    for name, scheme in config.config_groups.items():
        factory = TransformFactory.from_scheme(scheme, name=name)
        factory.apply_to_model(model)

    # attach config to model for compression/serialization
    setattr(model, TRANSFORM_CONFIG_NAME, config)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

"""
Utilities associated with offloading functionality

| ------------------------------------------------------------------------------------------------------ | # noqa: E501
| Operation  | Without offloading support             | With offloading support                          | # noqa: E501
| ---------- | -------------------------------------- | ------------------------------------------------ | # noqa: E501
| Update     | module.name.data.copy_(new_data)       | update_offload_parameter(module, name, new_data) | # noqa: E501
| ------------------------------------------------------------------------------------------------------ | # noqa: E501
"""

import contextlib
from typing import Literal

import torch
from compressed_tensors.offload import (
    align_module_device,
    align_modules,
    disable_offloading,
    get_execution_device,
    get_offloaded_device,
    offload_model,
    register_offload_module,
    remove_dispatch,
    update_offload_parameter,
)
from compressed_tensors.utils.helpers import deprecated


__all__ = [
    "get_execution_device",
    "get_offloaded_device",
    "update_parameter_data",
    "register_offload_parameter",
    "update_offload_parameter",
    "delete_offload_parameter",
    "has_offloaded_params",
    "disable_hf_hook",
    "disable_offload",
    "align_modules",
    "align_module_device",
    "register_offload_module",
    "delete_offload_module",
    "offloaded_dispatch",
    "disable_offloading",
    "remove_dispatch",
    "cast_to_device",
    "offload_to_weights_map",
    "delete_from_weights_map",
]


@deprecated("compressed_tensors.offload::update_offload_parameter")
def update_parameter_data(
    module: torch.nn.Module, new_param_data: torch.Tensor, param_name: str
):
    """
    Update the data of an existing parameter and its offload dict. Supports both
    parameters of offloaded modules and non-offloaded modules

    :param module: module containing the parameter to update
    :param new_param_data: tensor to update parameter with
    :param param_name: name of module parameter to update
    """
    update_offload_parameter(module, param_name, new_param_data)


""" Candidates for Upstreaming """


@deprecated()
def cast_to_device(device_spec: int | torch.device) -> torch.device:
    """
    Convert an integer device index or torch.device into a torch.device object.

    :param device_spec: Device index (int) or torch.device object.
                        Negative integers map to CPU.
    :return: torch.device corresponding to the given device specification.
    """
    if isinstance(device_spec, int):
        return torch.device(f"cuda:{device_spec}" if device_spec >= 0 else "cpu")
    return device_spec


@deprecated("module.register_parameter(name, parameter)")
def register_offload_parameter(
    module: torch.nn.Module,
    name: str,
    parameter: torch.nn.Parameter,
    offload_device: torch.device | Literal["disk"] | None = None,
):
    """
    Register a parameter to the given module which may be offloaded

    :param module: maybe offloaded module
    :param name: name of newly registered parameter
    :param parameter: parameter being registered
    :param offload_device: device on which weight will be offloaded to. If None is
        provided, then infer device from parameters on module
    """
    if offload_device == "disk":
        raise NotImplementedError("Disk offloading is not currently supported")

    module.register_parameter(name, parameter)


@deprecated("delattr(module, name)")
def delete_offload_parameter(module: torch.nn.Module, name: str):
    """
    Delete a parameter from a module which may be offloaded,
    including any metadata in _hf_hook

    :param module: maybe offloaded module
    :param name: name of parameter being deleted
    """
    delattr(module, name)


@deprecated("compressed_tensors.offload::unwrap_offload")
@contextlib.contextmanager
def disable_hf_hook(module: torch.nn.Module):
    raise ValueError()


@deprecated("delattr(base, name)")
def delete_offload_module(base: torch.nn.Module, name: str):
    """
    Delete a submodule from a model which may contain offloading
    :param base: parent module to delete submodule from
    :param name: name of submodule on parent
    """
    delattr(base, name)


@deprecated("compressed_tensors.offload::offload_model")
def offloaded_dispatch(
    module: torch.nn.Module,
    execution_device: torch.device,
    offload_device: torch.device | Literal["disk"] | None = None,
) -> torch.nn.Module:
    """
    Dispatch a model, keeping device parameters offloaded on their current device

    :param module: module containing parameters to offload
    :param execution_device: device that modules will onload and execute on
    :param offload_device: device that module parameters will offload to
    :return: module with offloading device hooks
    """
    if offload_device is not None:
        raise ValueError(
            "Passing offload_device to offloaded_dispatch is no longer supported"
        )
    offload_model(module, execution_device)


@deprecated("compressed_tensors.offload::align_module_device")
def disable_offload(module: torch.nn.Module):
    raise ValueError()


@deprecated()
def offload_to_weights_map(*args, **kwargs):
    raise ValueError()


@deprecated()
def delete_from_weights_map(*args, **kwargs):
    raise ValueError()


@deprecated()
def has_offloaded_params(module: torch.nn.Module) -> bool:
    """
    Checks if a module has offloaded parameters by checking if the given module has a
    AlignDevicesHook attached with offloading enabled

    Args:
        module (`torch.nn.Module`): The module to check for an offload hook.

    Returns:
        bool: `True` if the module has an offload hook and offloading is enabled,
        `False` otherwise.
    """
    return False



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import pytest


nbformat = pytest.importorskip("nbformat")
from nbconvert.preprocessors import ExecutePreprocessor  # noqa: E402


@pytest.mark.skip(
    reason="GHA not setup yet to run those tests. The test should work locally"
)
@pytest.mark.parametrize("notebook", ["examples/bitmask_compression.ipynb"])
def test_notebook_exec(notebook):
    with open(notebook) as f:
        nb = nbformat.read(f, as_version=4)
        ep = ExecutePreprocessor(timeout=600, kernel_name="python3")
        try:
            assert ep.preprocess(nb) is not None, f"Got empty notebook for {notebook}"
        except Exception:
            assert False, f"Failed executing {notebook}"



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import logging
import os
import re
from collections import defaultdict
from collections.abc import Generator, Iterable, Mapping
from typing import Iterator

import torch
from compressed_tensors.utils.internal import InternalModule


_LOGGER: logging.Logger = logging.getLogger(__name__)


__all__ = [
    "match_name",
    "match_named_modules",
    "match_named_parameters",
    "match_targets",
    "match_modules_set",
    "match_quantizable_tensors",
    "get_lowest_common_ancestor_name",
    "is_match",
    "is_narrow_match",
]


FusedMappping = Mapping[str, Iterable[str]]


def match_named_modules(
    model: torch.nn.Module,
    targets: Iterable[str] | None,
    ignore: Iterable[str] | None = None,
    fused: FusedMappping | None = None,
    warn_on_fail: bool = False,
) -> Generator[tuple[str, torch.nn.Module], None, None]:
    """
    Yields names and modules which match `targets` but do not match `ignore`.
    Values are returned in order of `model.named_modules()`

    :param model: model containing submodules to match against
    :param targets: target strings, potentially containing "re:" prefixes
    :param ignore: targets to ignore, potentially containing "re:" prefixes
    :fused: optional mapping from suffixes of fused modules to the suffixes of their
        corresponding shards. See `compressed_tensors.utils.match.is_match`
    :param warn_on_fail: if True, warns if any targets do not match any modules in model
    :return: generator of module names and modules
    """
    targets = targets or []
    ignore = ignore or []

    unmatched_targets = set(targets)

    for name, module in model.named_modules():
        for target in targets:
            if is_match(name, module, target, fused=fused):
                unmatched_targets -= {target}
                if not is_match(name, module, ignore, fused=fused):
                    yield name, module
                break

    if warn_on_fail:
        for target in unmatched_targets:
            _LOGGER.warning(
                f"Could not match `{target}` in instance of {model.__class__.__name__}"
            )


def match_named_parameters(
    model: torch.nn.Module,
    targets: Iterable[str] | None,
    ignore: Iterable[str] | None = None,
    fused: FusedMappping | None = None,
    warn_on_fail: bool = False,
) -> Generator[tuple[str, torch.nn.Module, torch.nn.Parameter], None, None]:
    """
    Yields parameters which match `targets` but do not match `ignore`.
    Values are returned in order of `model.named_modules()`

    :param model: model containing params to match against
    :param targets: target strings, potentially containing "re:" prefixes
    :param ignore: targets to ignore, potentially containing "re:" prefixes
    :fused: optional mapping from suffixes of fused modules to the suffixes of their
        corresponding shards. See `compressed_tensors.utils.match.is_match`
    :param warn_on_fail: if True, warns if any targets do not match any params in model
    :return: generator of fully-qualified param names, parent modules, and params
    """
    targets = targets or []
    ignore = ignore or []

    unmatched_targets = set(targets)
    for module_name, module in model.named_modules():
        if isinstance(module, InternalModule):
            continue

        for param_name, param in module.named_parameters(recurse=False):
            param_fqn = f"{module_name}.{param_name}"
            for target in targets:
                if match_name(param_fqn, target, fused):
                    unmatched_targets -= {target}

                    if not any(match_name(param_fqn, ign, fused) for ign in ignore):
                        yield param_fqn, module, param

    if warn_on_fail:
        for target in unmatched_targets:
            _LOGGER.warning(
                f"Could not match `{target}` in instance of {model.__class__.__name__}"
            )


def match_targets(
    name: str, module: torch.nn.Module, targets: Iterable[str] | None
) -> list[str]:
    """
    Returns the targets that match the given name and module.

    :param name: the name of the module
    :param module: the module to match
    :param targets: the target strings, potentially containing "re:" prefixes
    :return: the targets that match the given name and module

    Outputs are ordered by type: exact name match, regex name match, class name match
    """
    targets = targets or []

    if isinstance(module, InternalModule):
        return []

    # The order of the output `matches` list matters, they are arranged from most
    # specific to least specific, and this order will be used when merging configs.
    # The entries are sorted in the following order:
    #     1. matches on exact strings
    #     2. matches on regex patterns
    #     3. matches on module names (e.g. "Linear")

    targets = sorted(targets, key=lambda x: ("re:" in x, x))
    matched_targets = []
    for target in targets:
        if match_name(name, target):
            matched_targets.append(target)

    for target in targets:
        if _match_class(module, target) and target not in matched_targets:
            matched_targets.append(target)

    return matched_targets


def get_lowest_common_ancestor_name(names: list[str | None]) -> str:
    """
    Given a list of names, returns the lowest-scope common name ignoring Nones.

    Implementation is a small alteration of os.path.commonprefix
    https://docs.python.org/3/library/os.path.html#os.path.commonprefix

    ([s1, s2]->prefix->result)
    # case 0: multiple modules: [abc.a., abc.b.] -> .abc. -> abc
    # case 1: single module: [abc.] -> .abc. -> abc
    # case 2: substring modules: [abc., ab.] -> .ab -> ""
    # case 3: parent & child: [ab., ab.a.] -> .ab. -> ab
    """
    names = [name for name in names if name is not None]
    if len(names) == 0:
        return ""

    # 1) find longest shared prefix
    s1 = "." + min(names) + "."
    s2 = "." + max(names) + "."
    common_prefix = os.path.commonprefix([s1, s2])
    # 2) throw away right most dot and name fragment, throw away leftmost char
    # ".keep.thro" -> "keep", "." -> ""
    return common_prefix[1 : common_prefix.rfind(".")]


def match_modules_set(
    model: torch.nn.Module,
    targets: Iterable[str] | None,
    ignore: Iterable[str] | None = None,
    error_on_module_rematch: bool = True,
) -> Generator[list[list[torch.nn.Module]], None, None]:
    """
    Yields modules grouped by parent context.

    We group by parent context so that we can return ALL matches of a
    specific target that can be paired with another target. This is most
    relevant in the case of MoE modules with multiple modules for each
    expert i.e. post_attention_layernorm <-> mlp.expert.N.gate_proj,
    mlp.expert.N.up_proj for all N. The parent context will differ from
    one layer to another while being the same for one expert to another.

    Each returned group is a list (of lists) with the same size
    and order as `targets` while all matches for each target and
    the overall order of the groups are ordered in the same way
    as `model.named_modules`


    E.g. the following targets would yield modules belonging to the following layers:
    ```python3
    match_modules_set(model, ["q_proj", "k_proj", "v_proj"]) == (
        [
            [`layers.0.self_attn.q_proj`],
            [`layers.0.self_attn.k_proj`],
            [`layers.0.self_attn.v_proj`],
        ],
        [
            [`layers.1.self_attn.q_proj`],
            [`layers.1.self_attn.k_proj`],
            [`layers.1.self_attn.v_proj`],
        ],
        ...
    )
    ```

    This can be used to match layers to their corresponding downstream counterparts.
    For example, matching layer norms to their subsequent linear layers
    ```python3
    for norm, q, k, v in match_modules_set(model, (norm_tgt, q_tgt, k_tgt, v_tgt)):
        fuse_norm_linears(*norm, [*q, *k, *v])
    ```

    Alternatively for MoE you would get multiple matches
    per target per group, E.g.

    ```python3

    targets = [
        "post_attention_layernorm",
        "up_proj",
        "down_proj"
    ]
    match_modules_set(model, targets) == (
        [
            [layers.0.post_attention_layernorm],
            [
                `layers.0.mlp.experts.0.up_proj`,
                `layers.0.mlp.experts.1.up_proj`,
                ...
            ],
            [
                `layers.0.mlp.experts.0.down_proj`,
                `layers.0.mlp.experts.1.down_proj`,
                ...

            ]
        ], # <- first yield
        [
            [layers.1.post_attention_layernorm],
            [
                `layers.1.mlp.experts.0.up_proj`,
                `layers.1.mlp.experts.1.up_proj`,
                ...
            ],
            [
                `layers.1.mlp.experts.0.down_proj`,
                `layers.1.mlp.experts.1.down_proj`,
                ...
            ]
        ],
        ...
    )
    ```

    :param model: model containing modules to match against
    :param targets: target strings, potentially containing "re:" prefixes
    :param ignore: targets to ignore, potentially containing "re:" prefixes
    :param error_on_module_rematch: if True, errors when a module gets
      matched to multiple targets, if False, no error. (Defaults to True)
    """
    targets = targets or []
    ignore = ignore or []

    # as we iterate through modules and try to match them with targets,
    # the algorithm can be in 2 possible states:
    # 0) unmatched_targets > 0, i.e. some of the targets haven't been matched.
    #   Keep matching until all targets have at least one match
    # 1) unmatched_targets == 0 i.e. we have at least one match for each target.
    #   At this point we are unsure if we have a full set or if we need to add
    #   more matches.
    # There are 3 things that can happen once were in state 1:
    # A) found a new match with same parent_context,
    #   (add it to matches and keep going)
    # B) found a new match with different parent_context, i.e. we found a match
    #   that requires a deeper parent context, this indicates that this match
    #   should be part of a new set.
    #   (yield current set [not including newest match] and go back to state 0)
    # C) ran out of modules, we will always yield the final remaining set when
    #   we we've iterated through all the modules in the model.
    #   (yield final set then exit.)
    # Note: its possible to iterate through all the modules in the model while
    #   not having a full matched set if the user specified a bad matching, in
    #   that case something has gone wrong and we error
    matches = defaultdict(list)
    parent_context = None
    unmatched_targets = set(targets)

    for name, module in model.named_modules():
        matched_targets_for_cur_module = set()
        for target in targets:
            if is_match(name, module, target, ignore):
                new_parent_context = get_lowest_common_ancestor_name(
                    [name, parent_context]
                )

                # code for (B)
                if not unmatched_targets and new_parent_context != parent_context:
                    yield [matches[target] for target in targets]
                    matches = defaultdict(list)
                    new_parent_context = name
                    unmatched_targets = set(targets)

                matches[target].append(module)
                parent_context = new_parent_context
                unmatched_targets -= {target}
                matched_targets_for_cur_module |= {target}

        if len(matched_targets_for_cur_module) > 1 and error_on_module_rematch:
            raise ValueError(
                f"module: {name} was matched with multiple targets: "
                f"{matched_targets_for_cur_module} which is unexpected "
                "disable this check by setting `error_on_module_rematch = False`"
            )

    # never found anything
    if unmatched_targets == set(targets):
        return

    # code for (C)
    if not unmatched_targets:  # have a full matching
        yield [matches[target] for target in targets]
        return

    raise ValueError(
        f"Found a final incomplete set with matches found for keys: "
        f"{set(targets) - unmatched_targets} "
        f"but no matches found for keys: {unmatched_targets}"
    )


def is_match(
    name: str,
    module: torch.nn.Module,
    targets: str | Iterable[str],
    ignore: str | Iterable[str] = tuple(),
    fused: FusedMappping | None = None,
) -> bool:
    """
    Returns true if either module name or module parent classes match against target
    and the module is not an internal module. The name and module may refer to a fused
    module defined by vLLM. In these cases, a `fused` mapping must be provided.

    For example, in `vllm/model_executor/models/llama.py`:
    ```python
    packed_modules_mapping = {
        "qkv_proj": ["q_proj", "k_proj", "v_proj"],
        "gate_up_proj": ["gate_proj", "up_proj"]
    }
    ```

    :param name: name of module
    :param module: module to match
    :param target: target which matches name or module, potentially contains regex
    :fused: optional mapping from suffixes of fused modules to the suffixes of their
        corresponding shards
    """
    targets = [targets] if isinstance(targets, str) else targets
    ignore = [ignore] if isinstance(ignore, str) else ignore

    return not isinstance(module, InternalModule) and (
        any(
            match_name(name, target, fused) or _match_class(module, target)
            for target in targets
        )
        and not any(
            match_name(name, ign, fused) or _match_class(module, ign) for ign in ignore
        )
    )


def is_narrow_match(
    model: torch.nn.Module,
    targets: str | Iterable[str],
    name: str,
    module: torch.nn.Module | None = None,
) -> bool:
    """
    Checks if any of the targets narrowly match the module. A target narrowly matches
    a module if the target matches the module, but does not match the module's parent

    :param model: model containing both module and its parent
    :param targets: target strings, potentially containing "re:" prefixes
    :param name: name of module to match
    :param module: module to match. If none is provided, then get module from model
    :return: True if any of the targets narrow match the module
    """
    targets = [targets] if isinstance(targets, str) else targets
    module = module if module is not None else model.get_submodule(name)

    parent_name = name.rsplit(".", 1)[0]
    parent = model.get_submodule(parent_name)

    return any(
        is_match(name, module, target) and not is_match(parent_name, parent, target)
        for target in targets
    )


def match_name(name: str, target: str, fused: FusedMappping | None = None) -> bool:
    """
    Returns true if target string begins with "re:" and regex matches or if target
    string exactly matches name. If the name refers to a fused module defined by vLLM,
    a `fused` mapping must be provided.

    :param name: name of module
    :param target: target name, potentially contains regex
    :fused: optional mapping from suffixes of fused modules to the suffixes of their
        corresponding shards
    """
    if fused is not None:
        for fused_suffix in fused:
            if name.endswith(fused_suffix):
                name_stripped = name.removesuffix(fused_suffix)
                return any(
                    match_name(name_stripped + shard_suffix, target)
                    for shard_suffix in fused[fused_suffix]
                )

    if target.startswith("re:"):
        return re.match(target.removeprefix("re:"), name) is not None
    else:
        return target == name


def _match_class(module: torch.nn.Module, target: str) -> bool:
    """
    Returns true if any torch parent class names match the target string exactly.
    A special exception is made for vllm's `LinearBase` class which matches `Linear`

    :param module: module to match
    :param target: target which matches name or module
    """
    # will never match against a regex pattern since `:` is not allowed in class names
    return any(
        (
            issubclass(cls, torch.nn.Module)
            and (
                cls.__name__ == target
                or (cls.__name__ == "LinearBase" and target == "Linear")
            )
        )
        for cls in module.__class__.__mro__
    )


def match_quantizable_tensors(
    tensors: Mapping[str, torch.Tensor],
    ignore: Iterable[str],
    targets: Iterable[str] = tuple(),
    allow_nonquantizable: bool = False,
) -> Iterator[tuple[str, str]]:
    """
    Match all quantizable tensors that are not ignored and are
    targeted

    :param tensors: Mapping of name in safetensors file to actual tensor
    :param ignore: ignore individual tensor if any match is found with
        elements in ignore (regex allowed)
    :param targets: include if any match is found with elements in targets
        (regex allowed). Unlike model-based matching utils, this only checks
        names, not classes. If empty or if "Linear" is included, assume targets
        is all-inclusive.
    :param allow_nonquantizable: Override to include non-quantizable tensors,
        useful when performing other processing on tensors beyond quantization.

    :return: iterator of module_name and full tensor name meeting filtering
        criterion.
    """
    for name in list(tensors.keys()):
        module_name, param_name = name.rsplit(".", 1)

        is_quantizable = allow_nonquantizable or (
            (param_name == "weight") and not module_name.endswith("norm")
        )

        if len(targets) == 0 or "Linear" in targets:
            is_targeted = True
        else:
            is_targeted = any((match_name(module_name, target)) for target in targets)

        is_ignored = any(match_name(module_name, ign) for ign in ignore)

        if is_quantizable and is_targeted and not is_ignored:
            yield module_name, name



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

# flake8: noqa

from .base import *

# New per-format directories
from .dense import *
from .helpers import *
from .model_compressors import *
from .mxfp4 import *
from .mxfp8 import *
from .naive_quantized import *
from .nvfp4 import *
from .pack_quantized import *



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

# configs
QUANTIZATION_CONFIG_NAME = "quantization_config"
SPARSITY_CONFIG_NAME = "sparsity_config"
TRANSFORM_CONFIG_NAME = "transform_config"

# required fields
COMPRESSION_VERSION_NAME = "version"
QUANTIZATION_METHOD_NAME = "quant_method"
QUANTIZATION_METHOD = "compressed-tensors"



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import os

import pytest
import torch
from compressed_tensors.offload.cache.disk import DiskCache
from safetensors import safe_open
from tests.test_offload.cache.helpers import (
    _test_delete,
    _test_disable_offloading,
    _test_disable_onloading,
    _test_garbage_collect,
    _test_offload,
    _test_onload,
    _test_onloading,
    _test_shared_attributes,
    _test_tensor_subclass,
    _test_update_offload,
)
from tests.test_offload.conftest import assert_tensor_equal
from tests.testing_utils import requires_gpu


@pytest.fixture()
def onload_device():
    return torch.device("cuda")


@pytest.fixture()
def offload_device():
    return "disk"


@pytest.mark.unit
@requires_gpu
def test_delete(offload_device, onload_device, offload_cache):
    _test_delete(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu
def test_disable_offloading(offload_device, onload_device, offload_cache):
    _test_disable_offloading(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu
def test_disable_onloading(offload_device, onload_device, offload_cache):
    _test_disable_onloading(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu
def test_garbage_collect(offload_device, onload_device, offload_cache):
    _test_garbage_collect(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu
def test_offload(offload_device, onload_device, offload_cache):
    _test_offload(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu
@requires_gpu
def test_onload(offload_device, onload_device, offload_cache):
    _test_onload(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu
def test_onloading(offload_device, onload_device, offload_cache):
    _test_onloading(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu
def test_shared_attributes(offload_device, onload_device, offload_cache):
    _test_shared_attributes(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu
def test_tensor_subclass(offload_device, onload_device, offload_cache):
    _test_tensor_subclass(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu
def test_update_offload(offload_device, onload_device, offload_cache):
    _test_update_offload(offload_device, onload_device, offload_cache)


@pytest.mark.unit
def test_files(tmp_path):
    offload_dir = tmp_path / "offload_dir"
    os.mkdir(offload_dir)

    # initial write
    DiskCache.index = {}
    cache = DiskCache("cpu", offload_dir=str(offload_dir))
    tensor = torch.zeros(10)
    cache["weight"] = tensor

    files = os.listdir(offload_dir)
    assert len(DiskCache.index) == 1
    assert len(files) == 1
    with safe_open(offload_dir / files[0], framework="pt", device="cpu") as file:
        read_tensor = file.get_tensor("weight")
        assert_tensor_equal(read_tensor, tensor)

    # modify
    tensor = torch.ones(10)
    cache["weight"] = tensor

    files = os.listdir(offload_dir)
    assert len(DiskCache.index) == 1
    assert len(files) == 1
    with safe_open(offload_dir / files[0], framework="pt", device="cpu") as file:
        read_tensor = file.get_tensor("weight")
        assert_tensor_equal(read_tensor, tensor)

    # delete
    del cache["weight"]
    files = os.listdir(offload_dir)
    assert len(DiskCache.index) == 0
    assert len(files) == 0



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

# flake8: noqa
# isort: skip_file

from .forward import *
from .initialize import *
from .compressed import *
from .apply import *
from .helpers import *



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import pytest
import torch
from compressed_tensors.quantization import round_to_quantized_type_dtype
from compressed_tensors.quantization.quant_args import (
    QuantizationArgs,
    QuantizationStrategy,
    QuantizationType,
)
from compressed_tensors.quantization.utils import (
    generate_mx_scales,
    maybe_convert_from_mx_exp,
    round_to_power_2,
    should_generate_mx_scales,
)


def test_should_generate_mx_scales_mxfp8():
    """Test that should_generate_mx_scales returns True for MXFP8 args."""
    args = QuantizationArgs(
        num_bits=8,
        type=QuantizationType.FLOAT,
        strategy=QuantizationStrategy.GROUP,
        group_size=32,
        scale_dtype=torch.uint8,
        zp_dtype=torch.uint8,
    )
    assert should_generate_mx_scales(args) is True


def test_should_generate_mx_scales_mxfp4():
    """Test that should_generate_mx_scales returns True for MXFP4 args."""
    args = QuantizationArgs(
        num_bits=4,
        type=QuantizationType.FLOAT,
        strategy=QuantizationStrategy.GROUP,
        group_size=32,
        scale_dtype=torch.uint8,
        zp_dtype=torch.uint8,
    )
    assert should_generate_mx_scales(args) is True


def test_should_generate_mx_scales_regular_fp8():
    """Test that should_generate_mx_scales returns False for regular FP8."""
    args = QuantizationArgs(
        num_bits=8,
        type=QuantizationType.FLOAT,
        strategy=QuantizationStrategy.TENSOR,
    )
    assert should_generate_mx_scales(args) is False


def test_should_generate_mx_scales_wrong_group_size():
    """Test that should_generate_mx_scales returns False for non-32 group size."""
    args = QuantizationArgs(
        num_bits=8,
        type=QuantizationType.FLOAT,
        strategy=QuantizationStrategy.GROUP,
        group_size=128,
    )
    assert should_generate_mx_scales(args) is False


@pytest.mark.parametrize("dtype", [torch.bfloat16, torch.float32])
def test_mxfp8_scales_e2e(dtype):
    """End-to-end test for MXFP8 scale generation and conversion."""
    mock_weight = torch.normal(mean=0.0002, std=0.0576, size=(2880, 2880))

    x = mock_weight.reshape(*mock_weight.shape[:-1], -1, 32).to(dtype)
    min_vals = torch.amin(x, dim=-1)
    max_vals = torch.amax(x, dim=-1)

    min_vals = torch.min(min_vals, torch.zeros_like(min_vals))
    max_vals = torch.max(max_vals, torch.zeros_like(max_vals))
    block_max = torch.max(torch.abs(min_vals), torch.abs(max_vals))

    args = QuantizationArgs(
        num_bits=8,
        type=QuantizationType.FLOAT,
        strategy=QuantizationStrategy.GROUP,
        group_size=32,
        scale_dtype=torch.uint8,
        zp_dtype=torch.uint8,
    )

    scales = generate_mx_scales(block_max, num_bits=8)
    scales = round_to_quantized_type_dtype(scales, dtype=args.scale_dtype)

    converted_ct = maybe_convert_from_mx_exp(args=args, scale=scales)

    scales_exp = torch.log2(converted_ct)
    block_max_exp = torch.floor(torch.log2(round_to_power_2(block_max))) - 8
    assert torch.equal(scales_exp, block_max_exp)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import os

import pytest
import torch
from compressed_tensors.offload import disable_onloading, from_accelerate, to_accelerate
from compressed_tensors.offload.dist_utils import is_rank0
from tests.test_offload.conftest import torchrun
from tests.testing_utils import requires_gpu


acclerate = pytest.importorskip("accelerate")


def get_hf_dispatched_model(cuda_device, tmp_path):
    from accelerate.big_modeling import dispatch_model

    offload_dir = tmp_path / "offload_dir"
    os.mkdir(offload_dir)

    model = torch.nn.Sequential(
        torch.nn.Linear(5, 5), torch.nn.Linear(5, 5), torch.nn.Linear(5, 5)
    )

    if is_rank0():
        dispatch_model(
            model,
            {"0": 0, "1": "cpu", "2": "disk"},
            main_device=str(cuda_device),
            force_hooks=True,
            offload_dir=offload_dir,
        )
    else:
        model.to("meta")

    return model, offload_dir


@pytest.mark.unit
@requires_gpu
def test_conversion_lifecycle(cuda_device, tmp_path):
    model, offload_dir = get_hf_dispatched_model(cuda_device, tmp_path)

    exp_device_map = {
        "": (None, None),
        "0": (cuda_device, cuda_device),
        "1": (cuda_device, torch.device("cpu")),
        "2": (cuda_device, "disk"),
    }
    exp_hf_device_map = {"": "cpu", "0": str(cuda_device), "1": "cpu", "2": "disk"}

    # 1. from_accelerate (oneshot/ load_offloaded_model)
    device_map, _offload_dir = from_accelerate(model)
    assert device_map == exp_device_map
    with disable_onloading():
        state_dict = model.state_dict(keep_vars=True)

    # 2. to_accelerate (transformers save)
    hf_device_map = to_accelerate(model)
    assert hf_device_map == exp_hf_device_map

    # 3. from_accelerate (post-save restore)
    device_map, _offload_dir = from_accelerate(model)
    assert device_map == exp_device_map

    # Note that rank 0 tensor pointers remain unchanged:
    # - no extra gpu/cpu/disk memory is allocated
    # - disk index remains valid
    with disable_onloading():
        assert model.state_dict(keep_vars=True) == state_dict

    # 4. Just for completeness, test final invert (not part of lifecycle)
    hf_device_map = to_accelerate(model)
    assert hf_device_map == exp_hf_device_map


@pytest.mark.unit
@requires_gpu(2)
@torchrun(world_size=2)
def test_conversion_lifecycle_dist(cuda_device, tmp_path):
    test_conversion_lifecycle(cuda_device, tmp_path)
